# -*- coding: utf-8 -*-
# import io
import os
import string
import random
import sdk_test
from edge2client.constants import OFF


CURPATH = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))


class TestGlobalRule(sdk_test.TestSdk):
    def test_global_cert(self):
        key_file = os.path.join(CURPATH, 'tests', 'short_test.key')
        cert_file = os.path.join(CURPATH, 'tests', 'short_test.crt')
        with open(key_file) as f:
            key = f.read()
        with open(cert_file) as f:
            cert = f.read()

        cert_id = self.client.set_global_cert_key(key=key, cert=cert)
        self.assertIs(type(cert_id), int)
        self.assertGreater(cert_id, 0)

        data = self.client.get_global_cert_key(cert_id)
        self.assertEqual(data['domains'][0], 'test.com')

        key_file = os.path.join(CURPATH, 'tests', 'test.key')
        cert_file = os.path.join(CURPATH, 'tests', 'test.crt')
        with open(key_file) as f:
            key = f.read()
        with open(cert_file) as f:
            cert = f.read()

        ok = self.client.put_global_cert_key(cert_id=cert_id,
                                             key=key, cert=cert)
        self.assertTrue(ok)

        data = self.client.get_global_cert_key(cert_id)
        self.assertEqual(data['domains'][0], 'test.com')

        ok = self.client.del_global_cert_key(cert_id)
        self.assertTrue(ok)

        data = self.client.get_global_cert_key(cert_id)
        self.assertEqual(data, {})

    def test_global_user(self):
        app_id = self.app_id
        only_reader = ''.join([random.choice(string.ascii_letters)
                               for n in range(10)])
        pwd = ''.join([random.choice(string.ascii_letters + string.digits)
                       for n in range(12)])

        ok = self.client.add_global_user(
            name=only_reader, pwd=pwd, gid=[1])
        self.assertTrue(ok)

        ok = self.client.add_global_user(
            name='prefix_' + only_reader, pwd=pwd, gid=[1])
        self.assertTrue(ok)

        data = self.client.search_global_user('not_exist' + only_reader)
        user_id = data.get('id')
        self.assertEqual(user_id, None)

        data = self.client.search_global_user(only_reader)
        user_id = data.get('id')
        self.assertIs(type(user_id), int)
        self.assertGreater(user_id, 0)

        users = self.client.get_all_global_users()
        for global_user_id in users:
            self.assertIs(type(global_user_id), int)
            self.assertGreater(global_user_id, 0)

        user_id = self.client.add_app_user(name='not_exist' + only_reader)
        self.assertEqual(user_id, 'no "not_exist{}" user found'
                         .format(only_reader))

        user_id = self.client.add_app_user(name=only_reader)
        self.assertIs(type(user_id), int)
        self.assertGreater(user_id, 0)

        ok = self.client.put_app_user(id=user_id, name=only_reader,
                                      release=True)
        self.assertTrue(ok)

        data = self.client.get_app_user(id=user_id)
        self.assertTrue(data['id'], user_id)
        self.assertTrue(data['uid'], user_id)
        self.assertTrue(data['release'])
        self.assertTrue(data['write'])

        data = self.client.get_app_user(name=only_reader)
        self.assertTrue(data['id'], user_id)
        self.assertTrue(data['uid'], user_id)
        self.assertTrue(data['release'])
        self.assertTrue(data['write'])

        data = self.client.get_app_user(name='not_exist' + only_reader)
        self.assertEqual(data, {})

        data = self.client.get_app_user(user_id + 1)
        self.assertEqual(data, {})

        data = self.client.get_all_app_users()
        self.assertEqual(data[0]['username'], only_reader)
        self.assertTrue(data[0]['release'])
        self.assertTrue(data[0]['write'])

        ok = self.client.del_app_user(name=only_reader)
        self.assertTrue(ok)

        new_app_id = self.client.new_app(domains=['prefix' + self.apex],
                                         label=self.apex)

        ok = self.client.add_user_for_all_apps(name=only_reader)
        self.assertTrue(ok)

        data = self.client.get_all_app_users()
        self.assertEqual(data[0]['username'], only_reader)

        data = self.client.get_all_app_users(app_id=new_app_id)
        self.assertEqual(len(data), 1)
        self.assertEqual(data[0]['username'], only_reader)

        ok = self.client.del_app_user(name=only_reader)
        self.assertTrue(ok)

        data = self.client.get_all_app_users()
        self.assertEqual(len(data), 0)

        ok = self.client.del_app(new_app_id)
        self.assertTrue(ok)

        new_app_id = self.client.new_app(domains=['prefix' + self.apex],
                                         label=self.apex)
        self.client.use_app(new_app_id)
        ok = self.client.add_all_users_for_app()
        self.assertTrue(ok)

        data = self.client.get_all_app_users()
        self.assertGreater(len(data), 4)

        ok = self.client.del_app(new_app_id)
        self.assertTrue(ok)

        self.client.use_app(app_id)

    def test_global_rule(self):
        condition = [
            {'var': 'host', 'val': 'con.' + self.apex},
            {'var': ['req-header', 'Referer'],
             'vals': [[r'foo\d+', 'rx'], 'foo.com']}
        ]
        conseq = {
            'enable-websocket': {},
            'redirect': {'url': '/cn/2017/', 'code': 302}
        }
        rule_id = self.client.new_global_rule(condition=condition,
                                              conseq=conseq)
        self.assertIs(type(rule_id), int)
        self.assertGreater(rule_id, 0)

        data = self.client.get_global_rule(rule_id)
        if data['actions'][0]['type'] == 'redirect':
            self.assertEqual(data['actions'][0]['redirect']['url'],
                             '/cn/2017/')
            self.assertEqual(data['actions'][1]['type'], 'enable-websocket')
        else:
            self.assertEqual(data['actions'][1]['redirect']['url'],
                             '/cn/2017/')
            self.assertEqual(data['actions'][0]['type'],
                             'enable-websocket')

        conseq['redirect']['url'] = '/cn/2018/'
        ok = self.client.put_global_rule(rule_id=rule_id, condition=condition,
                                         conseq=conseq)
        self.assertTrue(ok)
        data = self.client.get_global_rule(rule_id)
        if data['actions'][0]['type'] == 'redirect':
            self.assertEqual(data['actions'][0]['redirect']['url'],
                             '/cn/2018/')
        else:
            self.assertEqual(data['actions'][1]['redirect']['url'],
                             '/cn/2018/')

        # turn off condition
        ok = self.client.put_global_rule(rule_id=rule_id, condition=OFF,
                                         conseq=conseq)
        self.assertTrue(ok)

        data = self.client.get_global_rule(rule_id)
        self.assertEqual(data.get('conditions', None), None)

        data = self.client.get_all_global_rules()
        self.assertEqual(len(data), 1)
        data = data[0]
        if data['actions'][0]['type'] == 'redirect':
            self.assertEqual(data['actions'][0]['redirect']['url'],
                             '/cn/2018/')
        else:
            self.assertEqual(data['actions'][1]['redirect']['url'],
                             '/cn/2018/')

        ok = self.client.del_global_rule(rule_id)
        self.assertTrue(ok)
        data = self.client.get_global_rule(rule_id)
        self.assertEqual(data, {})

    def test_global_var(self):
        random_name = ''.join([random.choice(string.ascii_letters)
                               for n in range(10)])
        name = 'is-whitelist-' + random_name
        var_id = self.client.new_global_var(name=name, var_type='string',
                                            default='no')
        self.assertIs(type(var_id), int)
        self.assertGreater(var_id, 0)

        condition = [
            {'global_var': var_id, 'vals': ['no']}
        ]
        conseq = {
            'limit-req-rate': {'limit_key': 'uri', 'rate_shape': 1,
                               'rate_reject': 1, 'rate_shape_unit': 'r/min',
                               'rate_reject_unit': 'r/min'}
        }
        rule_id = self.client.new_global_rule(condition=condition,
                                              conseq=conseq)
        self.assertIs(type(rule_id), int)
        self.assertGreater(rule_id, 0)

        ok = self.client.put_global_var(var_id=var_id, name=name,
                                        var_type='string', default='yes')
        self.assertTrue(ok)

        data = self.client.get_global_var(var_id)
        self.assertEqual(data['default'], 'yes')
        self.assertEqual(data['name'], name)

        ok = self.client.del_global_rule(rule_id)
        self.assertTrue(ok)

        ok = self.client.del_global_var(var_id)
        self.assertTrue(ok)

    def test_global_ngx_config(self):
        config = {'keepalive_timeout': 10, 'enable_open_file_cache': False}
        ok = self.client.set_global_ngx_config(config)
        self.assertTrue(ok)

        data = self.client.get_global_ngx_config()
        self.assertEqual(data['keepalive_timeout'], 10)
        self.assertFalse(data['enable_open_file_cache'])

    def test_request_id(self):
        ok = self.client.disable_request_id()
        self.assertTrue(ok)

        status = self.client.get_request_id_status()
        self.assertFalse(status)

        ok = self.client.enable_request_id()
        self.assertTrue(ok)

        status = self.client.get_request_id_status()
        self.assertTrue(status)

        data = self.client.get_global_misc_config()
        self.assertTrue(data['enabled_req_id'])

        ok = self.client.disable_request_id()
        self.assertTrue(ok)

        status = self.client.get_request_id_status()
        self.assertFalse(status)

        ok = self.client.set_global_misc_config({'enabled_req_id': True})
        self.assertTrue(ok)

        data = self.client.get_global_misc_config()
        self.assertTrue(data['enabled_req_id'])

        ok = self.client.disable_request_id()
        self.assertTrue(ok)

    def test_global_waf_rule(self):
        data = self.client.get_all_global_waf_rules()
        for rule in data:
            rule_id = rule['id']
            if rule_id > 10000:
                self.client.del_global_waf_rule(rule_id)

        # add global user defined waf user
        waf_rule_id = self.client.new_global_waf_rule(
            name='foo',
            code='''uri-arg('foo') =>
    waf-mark-risk(level: 'definite', msg: 'found foo');''')
        self.assertIs(type(waf_rule_id), int)
        self.assertGreater(waf_rule_id, 0)

        ok = self.client.put_global_waf_rule(
            rule_id=waf_rule_id, name='bar',
            code='''uri-arg('bar') =>
    waf-mark-risk(level: 'definite', msg: 'found bar');''')
        self.assertTrue(ok)

        data = self.client.get_global_waf_rule(waf_rule_id)
        self.assertEqual(data['name'], 'bar')
        self.assertEqual(data['code'], '''uri-arg('bar') =>
    waf-mark-risk(level: 'definite', msg: 'found bar');''')

        # add new rewrite rule without user defined waf rules
        condition = [{'var': 'uri', 'op': 'prefix', 'val': '/foo'}]
        waf_rule = {'rule_sets': [7,8,9,10,11,12,13,14,15,16,17],
                    'action': '403 Forbidden',
                    'threshold': 'low'}
        rule_id = self.client.new_rule(condition=condition,
                                       waf=waf_rule, last=True)
        self.assertIs(type(rule_id), int)
        self.assertGreater(rule_id, 0)

        data = self.client.get_rule(rule_id)
        self.assertIn(7, data['waf']['rule_sets'])
        self.assertIn(8, data['waf']['rule_sets'])
        self.assertNotIn(waf_rule_id, data['waf']['rule_sets'])

        # update rewrite rule with user defined waf rules
        waf_rule = {'rule_sets': [waf_rule_id], 'action': '403 Forbidden',
                    'threshold': 'high'}
        ok = self.client.put_rule(rule_id=rule_id,
                                  waf=waf_rule, condition=condition)
        self.assertTrue(ok)

        data = self.client.get_rule(rule_id)
        # self.assertIn(2, data['waf']['rule_sets'])
        # self.assertIn(3, data['waf']['rule_sets'])
        self.assertIn(waf_rule_id, data['waf']['rule_sets'])
        self.assertEqual(data['waf']['threshold_score'], 1000)

        ok = self.client.del_rule(rule_id)
        self.assertTrue(ok)

        ok = self.client.del_global_waf_rule(waf_rule_id)
        self.assertTrue(ok)

    def test_req_rewrite_2(self):
        condition = [
            {'var': 'host', 'val': 'con.' + self.apex},
            {'var': ['req-header', 'Referer'],
             'vals': [[r'foo\d+', 'rx'], 'foo.com']}
        ]
        conseq = [
            {'set-proxy-header': {
                'header': 'Foo',
                'value': 'default',
            }},
            {'set-proxy-header': {
                'header': 'Host',
                'value': "orig",
            }},
            {'print': {
                'msg': 'hello'
            }}
        ]
        rule_id = self.client.new_rule(condition=condition, conseq=conseq)
        self.assertIs(type(rule_id), int)
        self.assertGreater(rule_id, 0)

        data = self.client.get_rule(rule_id)
        self.assertEqual(len(data['actions']), 3)
        self.assertEqual(data['actions'][0]['set-proxy-header'],
                         {'header': 'Foo', 'value': 'default'})
        self.assertEqual(data['actions'][1]['set-proxy-header'],
                         {'header': 'Host', 'value': 'orig'})
        self.assertEqual(data['actions'][2]['print'], {'msg': 'hello'})

        ok = self.client.del_rule(rule_id)
        self.assertTrue(ok)

    def test_global_action(self):
        name = 'action_' + self.apex
        condition = [
            {'var': 'req-method', 'vals': ['GET', 'POST', 'HEAD']}
        ]
        conseq = [
            {
                'exit': {'code': 403}
            }
        ]
        action_id = self.client.new_global_action(name=name,
                                                  condition=condition,
                                                  conseq=conseq)
        self.assertIs(type(action_id), int)
        self.assertGreater(action_id, 0)

        data = self.client.get_global_action(action_id)
        self.assertEqual(data['conditions'][0]['variable'],
                         {'name': 'req-method'})
        self.assertEqual(data['conditions'][0]['values'][0]['val'], 'GET')
        self.assertEqual(data['conditions'][0]['values'][1]['val'], 'POST')
        self.assertEqual(data['conditions'][0]['values'][2]['val'], 'HEAD')
        self.assertEqual(data['actions'][0]['exit'], {'code': 403})

        conseq = [
            {'exit': {'code': 404}}
        ]
        ok = self.client.put_global_action(name=name, action_id=action_id,
                                           condition=condition,
                                           conseq=conseq)
        self.assertTrue(ok)

        data = self.client.get_global_action(action_id)
        self.assertEqual(data['actions'][0]['exit'], {'code': 404})

        conseq = [
            {
                'global_action': action_id
            }
        ]
        rule_id = self.client.new_rule(conseq=conseq, top=1)
        self.assertIs(type(rule_id), int)
        self.assertGreater(rule_id, 0)

        data = self.client.get_rule(rule_id)
        self.assertEqual(data['actions'][0]['global_action_id'], action_id)

        ok = self.client.del_rule(rule_id)
        self.assertTrue(ok)

        data = self.client.get_all_global_actions()
        self.assertEqual(data[0]['name'], name)

        ok = self.client.del_global_action(action_id)
        self.assertTrue(ok)

        data = self.client.get_global_action(action_id)
        self.assertEqual(data, {})

    def test_get_all_global_waf_rules(self):
        data = self.client.get_all_global_waf_rules()
        for rule in data:
            rule_id = rule['id']
            if rule_id > 10000:
                self.client.del_global_waf_rule(rule_id)

        waf_rule_foo_id = self.client.new_global_waf_rule(
            name='foo',
            code='''uri-arg('foo') =>
    waf-mark-risk(level: 'definite', msg: 'found foo');''')
        self.assertIs(type(waf_rule_foo_id), int)
        self.assertGreater(waf_rule_foo_id, 0)

        waf_rule_bar_id = self.client.new_global_waf_rule(
            name='bar',
            code='''uri-arg('bar') =>
    waf-mark-risk(level: 'definite', msg: 'found bar');''')
        self.assertIs(type(waf_rule_bar_id), int)
        self.assertGreater(waf_rule_bar_id, 0)

        data = self.client.get_all_global_waf_rules(detail=True)
        self.assertEqual(len(data), 14)

        self.assertEqual(data[-1]['name'], 'bar')
        self.assertEqual(data[-1]['code'], '''uri-arg('bar') =>
    waf-mark-risk(level: 'definite', msg: 'found bar');''')
        self.assertEqual(data[-2]['name'], 'foo')
        self.assertEqual(data[-2]['code'], '''uri-arg('foo') =>
    waf-mark-risk(level: 'definite', msg: 'found foo');''')

        data = self.client.get_all_global_waf_rules()

        self.assertEqual(len(data), 14)

        self.assertEqual(data[-1]['name'], 'bar')
        self.assertEqual(data[-1].get('code', None), None)
        self.assertEqual(data[-2]['name'], 'foo')
        self.assertEqual(data[-2].get('code', None), None)

        ok = self.client.del_global_waf_rule(waf_rule_foo_id)
        self.assertTrue(ok)

        ok = self.client.del_global_waf_rule(waf_rule_bar_id)
        self.assertTrue(ok)

        data = self.client.get_all_global_waf_rules()
        self.assertEqual(len(data), 12)

    def test_global_cert_adn_app_cert(self):
        key_file = os.path.join(CURPATH, 'tests', 'test.key')
        cert_file = os.path.join(CURPATH, 'tests', 'test.crt')
        with open(key_file) as f:
            key = f.read()
        with open(cert_file) as f:
            cert = f.read()

        global_cert_id = self.client.set_global_cert_key(key=key, cert=cert)

        data = self.client.get_global_cert_key(global_cert_id)
        self.assertEqual(data['domains'][0], 'test.com')

        new_app_id = self.client.new_app(domains=['test.com'], label='test')
        self.assertIs(type(new_app_id), int)
        self.assertGreater(new_app_id, 0)

        cert_id = self.client.set_cert_key(global_cert_id=global_cert_id)
        self.assertIs(type(cert_id), int)
        self.assertGreater(cert_id, 0)

        ok = self.client.del_cert_key(cert_id)
        self.assertTrue(ok)

        ok = self.client.del_global_cert_key(global_cert_id)
        self.assertTrue(ok)

        data = self.client.get_global_cert_key(global_cert_id)
        self.assertEqual(data, {})

        ok = self.client.del_app(new_app_id)
        self.assertTrue(ok)

        self.client.use_app(self.app_id)

    def test_static_file(self):
        data = self.client.get_all_static_files()
        for file in data:
            file_id = file['id']
            self.client.del_static_file(file_id)

        file_id = self.client.upload_static_file(
            filename="500.html", content='content', label='500.html')
        self.assertIs(type(file_id), int)
        self.assertGreater(file_id, 0)

        data = self.client.get_static_file(file_id)
        self.assertEqual(data['label'], '500.html')
        self.assertEqual(data['type'], 'file')
        self.assertEqual(data['id'], file_id)

        condition = [{'var': 'uri', 'op': 'eq', 'val': '/test'}]
        conseq = {
            'set-error-page': {
                'status': [500, 501, 502],
                'content_type': 'text/html; charset=utf8',
                'file_id': file_id
            }
        }
        rule_id = self.client.new_rule(condition=condition, conseq=conseq)
        self.assertIs(type(rule_id), int)
        self.assertGreater(rule_id, 0)

        data = self.client.get_rule(rule_id)
        self.assertEqual(data['actions'][0]['set-error-page']['status'],
                         [500, 501, 502])
        self.assertEqual(data['actions'][0]['set-error-page']['file_id'],
                         file_id)

        ok = self.client.del_rule(rule_id)
        self.assertTrue(ok)

        new_file_id = self.client.upload_static_file(
            filename='400.html', content='content', label='400.html', gid=[2,3])
        self.assertIs(type(new_file_id), int)
        self.assertGreater(new_file_id, 0)

        data = self.client.get_all_static_files()
        found = False
        for f in data:
            if f['id'] == file_id:
                found = True
                self.assertEqual(f['label'], '500.html')

            if f['id'] == new_file_id:
                found = True
                self.assertEqual(f['label'], '400.html')
                self.assertEqual(f['gid'], [2,3])

        self.assertEqual(found, True)

        ok = self.client.del_static_file(file_id)
        self.assertTrue(ok)
        ok = self.client.del_static_file(new_file_id)
        self.assertTrue(ok)

    def test_global_upstream(self):
        up_id = self.client.new_global_upstream(
            name='origin-upstream',
            servers=[
                {'ip': '172.22.31.1', 'port': 80},
                {'ip': '172.22.31.2', 'port': 80, 'weight': 2}
            ])
        self.assertIs(type(up_id), int)
        self.assertGreater(up_id, 0)

        ok = self.client.put_global_upstream(
            up_id=up_id,
            name='test-upstream',
            servers=[
                {'ip': '172.22.31.1', 'port': 80}
            ])
        self.assertTrue(ok)

        data = self.client.get_global_upstream(up_id)
        self.assertEqual(data['name'], 'test-upstream')
        self.assertEqual(len(data['nodes']), 1)
        self.assertEqual(data['nodes'][0]['ip'], '172.22.31.1')

        data = self.client.get_all_global_upstreams()
        self.assertEqual(len(data), 1)
        self.assertEqual(data[0]['name'], 'test-upstream')

        ok = self.client.del_global_upstream(up_id)
        self.assertTrue(ok)

        data = self.client.get_global_upstream(up_id)
        self.assertEqual(data, {})

    def test_global_upstream_and_proxy(self):
        up_id = self.client.new_global_upstream(
            name='origin-upstream',
            servers=[
                {'domain': 'test.com', 'port': 80},
                {'ip': '172.22.31.2', 'port': 80, 'weight': 2}
            ])
        self.assertIs(type(up_id), int)
        self.assertGreater(up_id, 0)

        proxy_rule = {
            'upstreams': [{'global_upstream': up_id, 'weight': 2}],
            'retries': 2,
            "retry_condition": ["invalid_header", "http_500"],
            "balancer_algorithm": "hash",
            "balancer_vars": [
                {"name": "uri"},
                {
                    "name": "uri-arg",
                    "args": "foo"
                }
            ]
        }
        rule_id = self.client.new_rule(proxy=proxy_rule, top=1)
        self.assertIs(type(rule_id), int)
        self.assertGreater(rule_id, 0)

        condition = [
            {'var': 'host', 'val': 'con.' + self.apex},
            {'var': ['req-header', 'Referer'],
             'vals': [[r'foo\d+', 'rx'], 'foo.com']}
        ]
        conseq = {'redirect': {'url': '/cn/2017/', 'code': 302}}

        proxy_rule['upstreams'][0]['weight'] = 3
        ok = self.client.put_rule(
            rule_id=rule_id, condition=condition,
            conseq=conseq, proxy=proxy_rule)
        self.assertTrue(ok)

        proxy_rule['retries'] = -1
        ok = self.client.put_proxy_rule(rule_id=rule_id, proxy=proxy_rule)
        self.assertTrue(ok)

        data = self.client.get_rule(rule_id)
        self.assertEqual(data['actions'][0]['redirect']['url'], '/cn/2017/')
        self.assertEqual(data['proxy']['upstream'][0]['global_cluster'], up_id)
        self.assertEqual(data['proxy']['upstream'][0]['weight'], 3)
        self.assertEqual(data['proxy']['retries'], -1)
        self.assertEqual(data['proxy']['retry_condition'],
                         ["invalid_header", "http_500"])
        self.assertEqual(data['proxy']['balancer']['algorithm'], 'hash')
        self.assertEqual(data['proxy']['balancer']['variables'][0]['name'],
                         'uri')
        self.assertEqual(data['proxy']['balancer']['variables'][1]['name'],
                         'uri-arg')
        self.assertEqual(data['proxy']['balancer']['variables'][1]['args'],
                         'foo')

        ok = self.client.del_rule(rule_id)
        self.assertTrue(ok)

        data = self.client.get_rule(rule_id)
        self.assertEqual(data, {})

        ok = self.client.del_global_upstream(up_id)
        self.assertTrue(ok)

        data = self.client.get_global_upstream(up_id)
        self.assertEqual(data, {})

    def test_waf_whitelist(self):
        condition = [
            {'var': 'host', 'val': 'con.' + self.apex},
            {'var': ['req-header', 'Referer'],
             'vals': [[r'foo\d+', 'rx'], 'foo.com']}
        ]

        waf_id = self.client.new_waf_whitelist(condition=condition,
                                               rule_sets=[7,8,9,10,11,12,13,14,15,16,17])
        self.assertIs(type(waf_id), int)
        self.assertGreater(waf_id, 0)

        data = self.client.get_waf_whitelist(waf_id)
        self.assertEqual(data['rule_sets'], [7,8,9,10,11,12,13,14,15,16,17])
        self.assertEqual(data['conditions'][0]['variable']['name'], 'host')

        condition = [{
            'var': 'uri',
            'op': '!contains',
            'vals': [
                ['/123.$', 'rx'],
                ['3333ter/', 'rx']
            ]
        }]
        ok = self.client.put_waf_whitelist(whitelist_id=waf_id,
                                           condition=condition,
                                           rule_sets=[7,8,9,10,11,12,13,14,15,16,17])
        self.assertTrue(ok)
        data = self.client.get_waf_whitelist(waf_id)
        self.assertEqual(data['rule_sets'], [7,8,9,10,11,12,13,14,15,16,17])

        data = self.client.get_all_waf_whitelists()
        self.assertEqual(len(data), 1)
        data = data[0]
        self.assertEqual(data['rule_sets'], [7,8,9,10,11,12,13,14,15,16,17])

        ok = self.client.del_waf_whitelist(waf_id)
        self.assertTrue(ok)
        data = self.client.get_all_waf_whitelists()
        self.assertEqual(data, [])



# -*- coding: utf-8 -*-
# import io
import sdk_test

class TestAppMetrics(sdk_test.TestSdk):
    def test_app_metrics(self):
        data = self.client.get_app_metrics(self.app_id)
        self.assertEqual(data, [])



# -*- coding: utf-8 -*-
# import io
import sdk_test


class TestUpstream(sdk_test.TestSdk):
    def test_upstream_and_proxy(self):
        up_id = self.client.new_upstream(
            name='origin-upstream',
            servers=[
                {'domain': 'test.com', 'port': 80},
                {'ip': '172.22.31.2', 'port': 80, 'weight': 2}
            ])
        self.assertIs(type(up_id), int)
        self.assertGreater(up_id, 0)

        changes = self.client.pending_changes()
        self.assertEqual(changes, 1)

        upstream = self.client.get_upstream(up_id)
        server1_id1 = upstream['nodes'][0]['id']

        ok = self.client.put_upstream(
            up_id, name='origin-upstream',
            servers=[
                {'domain': 'app.com', 'port': 8080},
                {'ip': '172.22.31.2', 'port': 8080, 'weight': 1}
            ])

        upstream = self.client.get_upstream(up_id)
        server1_id2 = upstream['nodes'][0]['id']

        self.assertNotEqual(server1_id1, server1_id2)

        ok = self.client.put_upstream(
            up_id, name='origin-upstream',
            servers=[
                {'domain': 'app.com', 'port': 8080, 'id': server1_id2},
                {'ip': '172.22.31.2', 'port': 8080, 'weight': 1}
            ])

        upstream = self.client.get_upstream(up_id)
        server1_id3 = upstream['nodes'][0]['id']

        self.assertEqual(server1_id2, server1_id3)

        ok = self.client.put_upstream(
            up_id, name='origin-upstream',
            servers=[
                {'domain': 'app.com', 'port': 8080},
                {'ip': '172.22.31.2', 'port': 8080, 'weight': 1}
            ],
            health_checker={
                'http_req_uri': '/status',
                'http_req_host': 'test.com',
                'interval': 3,
                'interval_unit': 'sec',
                'timeout': 1,
                'fall': 3,
                'rise': 2,
                'valid_statuses': [200, 302],
                'report_interval': 3,
                'report_interval_unit': 'min'
            }
        )
        self.assertTrue(ok)

        data = self.client.get_upstream(up_id)
        self.__assert_upstream(data, {
            'name': 'origin-upstream',
            'nodes': [
                {'domain': 'app.com', 'port': 8080, 'status': 1},
                {'ip': '172.22.31.2', 'weight': 1, 'status': 1}
            ],
            'checker': {'interval': 3, 'rise': 2, 'http_req_uri': '/status'}
        })

        upstreams = self.client.get_all_upstreams()
        self.assertTrue('origin-upstream' in upstreams.keys())
        self.assertIs(type(upstreams['origin-upstream']), int)

        changes = self.client.pending_changes()
        self.assertEqual(changes, 4)

        self.client.new_release()
        changes = self.client.pending_changes()
        self.assertEqual(changes, 0)

        backup_up_id = self.client.new_upstream(
            name='backup-upstream',
            servers=[
                {'ip': '172.22.31.3', 'port': 80},
                {'ip': '172.22.31.4', 'port': 80, 'weight': 3}
            ])
        self.assertIs(type(backup_up_id), int)
        self.assertGreater(backup_up_id, up_id)

        proxy_rule = {
            'upstreams': [],
            'backup_upstreams': [],
            'upstream_el_code': 'true => set-upstream-addr(ip: "127.0.0.1", port: 80);',
            'retries': 2,
            "retry_condition": ["invalid_header", "http_500"],
            "balancer_algorithm": "hash",
            "balancer_vars": [
                {"name": "uri"},
                {
                    "name": "uri-arg",
                    "args": "foo"
                }
            ]
        }

        rule_id = self.client.new_rule(proxy=proxy_rule, top=1)
        self.assertIs(type(rule_id), int)
        self.assertGreater(rule_id, 0)

        data = self.client.get_rule(rule_id)
        self.__assert_proxy_rule(data, {
            'proxy': {
                'upstream_el_code': 'true => set-upstream-addr(ip: "127.0.0.1", port: 80);',
                'retries': 2,
                'retry_condition': ['invalid_header', 'http_500'],
                'balancer': {
                    'algorithm': 'hash',
                    'variables': [
                        {'name': 'uri'},
                        {'name': 'uri-arg', 'args': 'foo'}
                    ]
                }
            }
        })

        proxy_rule = {
            'upstreams': [{'upstream': up_id, 'weight': 2}],
            'backup_upstreams': [{'upstream': backup_up_id}],
            'retries': 2,
            "retry_condition": ["invalid_header", "http_500"],
            "balancer_algorithm": "hash",
            "balancer_vars": [
                {"name": "uri"},
                {
                    "name": "uri-arg",
                    "args": "foo"
                }
            ]
        }
        ok = self.client.put_rule(rule_id=rule_id, proxy=proxy_rule)
        self.assertIs(type(rule_id), int)
        self.assertGreater(rule_id, 0)

        proxy_rule['upstreams'][0]['weight'] = 3
        ok = self.client.put_rule(rule_id=rule_id, proxy=proxy_rule)
        self.assertTrue(ok)

        data = self.client.get_rule(rule_id)

        data['proxy']['retries'] = 111
        data['proxy']['upstream'][0]['weight'] = 9
        data['proxy']['backup_upstream'][0]['weight'] = 9
        data['proxy']['upstream'][0]['upstream'] = data['proxy']['upstream'][0]['cluster']
        data['proxy']['backup_upstream'][0]['upstream'] = data['proxy']['upstream'][0]['cluster']
        ok = self.client.put_rule(rule_id=rule_id, proxy=data.get('proxy'))
        self.assertTrue(ok)

        self.__assert_proxy_rule(data, {
            'proxy': {
                'upstream': [{
                    'cluster': up_id,
                    'weight': 9,
                }],
                'backup_upstream': [{
                    'cluster': backup_up_id,
                    'weight': 9,
                }],
                'upstream_el_code': '',
                'retries': 111,
                'retry_condition': ['invalid_header', 'http_500'],
                'balancer': {
                    'algorithm': 'hash',
                    'variables': [
                        {'name': 'uri'},
                        {'name': 'uri-arg', 'args': 'foo'}
                    ]
                }
            }
        })

        data = self.client.get_all_rules_by_upstream_ip('172.22.31.2')
        self.assertEqual(len(data[self.app_id]), 1)
        self.assertEqual(data[self.app_id][0]['proxy']['retries'], 111)

        ok = self.client.del_rule(rule_id)
        self.assertTrue(ok)

        data = self.client.get_rule(rule_id)
        self.assertEqual(data, {})

        ok = self.client.del_upstream(up_id)
        self.assertTrue(ok)

        ok = self.client.del_upstream(backup_up_id)
        self.assertTrue(ok)

        data = self.client.get_upstream(backup_up_id)
        self.assertEqual(data, {})

    def __assert_upstream(self, upstream, expected):
        self.assertEqual(upstream['name'], expected['name'])

        for i, node in enumerate(upstream['nodes']):
            expected_node = expected['nodes'][i]
            for item in ('domain', 'port', 'status', 'ip', 'weight'):
                if expected_node.get(item):
                    self.assertEqual(node.get(item), expected_node.get(item))

        upstream_checker = upstream['checker']
        expected_checker = expected['checker']
        for item in ('interval', 'rise', 'http_req_uri'):
            self.assertEqual(upstream_checker.get(item),
                             expected_checker.get(item))

    def __assert_proxy_rule(self, proxy, expected):
        if 'upstream_el_code' in proxy['proxy']:
            self.assertEqual(proxy['proxy']['upstream_el_code'],
                        expected['proxy']['upstream_el_code'])
        else:
            for upstream_keyword in ('upstream', 'backup_upstream'):
                proxy_upstreams = proxy['proxy'][upstream_keyword]
                expected_upstreams = expected['proxy'][upstream_keyword]
                for i, proxy_upstream in enumerate(proxy_upstreams):
                    expected_upstream = expected_upstreams[i]
                    for item in ('cluster', 'weight'):
                        if expected_upstream.get(item):
                            self.assertEqual(proxy_upstream.get(item),
                                            expected_upstream.get(item))

        self.assertEqual(proxy['proxy']['retries'],
                         expected['proxy']['retries'])
        self.assertEqual(proxy['proxy']['retry_condition'],
                         expected['proxy']['retry_condition'])
        self.assertEqual(proxy['proxy']['balancer']['algorithm'],
                         expected['proxy']['balancer']['algorithm'])

        variables = proxy['proxy']['balancer']['variables']
        expected_vars = expected['proxy']['balancer']['variables']

        for i, variable in enumerate(variables):
            expected_var = expected_vars[i]
            for item in ('name', 'args'):
                if expected_var.get(item):
                    self.assertEqual(variable.get(item),
                                     expected_var.get(item))



# -*- coding: utf-8 -*-
# import io
import sdk_test


class TestAppBasicAuth(sdk_test.TestSdk):
    def test_app_basic_auth(self):
        app_id = self.app_id

        data = self.client.get_all_app_basic_auth_user_groups()
        self.assertEqual(len(data), 0)

        data = self.client.get_all_app_basic_auth_user_groups(app_id)
        self.assertEqual(len(data), 0)

        ug_id1 = self.client.new_app_basic_auth_user_group("test1")
        ug_id2 = self.client.new_app_basic_auth_user_group("test2", app_id=app_id)
        data = self.client.get_all_app_basic_auth_user_groups()
        self.assertEqual(len(data), 2)

        ok = self.client.del_app_basic_auth_user_group(ug_id2)
        self.assertEqual(ok, True)

        data = self.client.get_all_app_basic_auth_user_groups()
        self.assertEqual(len(data), 1)

        ok = self.client.put_app_basic_auth_user_group(ug_id1, name="new-test1", label="new-test1-label")
        self.assertEqual(ok, True)

        ok = self.client.put_app_basic_auth_user_group(ug_id1, name="new-test1", label="new-test1-label", app_id=app_id)
        self.assertEqual(ok, True)

        data = self.client.get_app_basic_auth_user_group(ug_id1)
        self.assertEqual(data['name'], "new-test1")
        self.assertEqual(data['label'], "new-test1-label")

        data = self.client.get_app_basic_auth_user_group(ug_id1, app_id)
        self.assertEqual(data['name'], "new-test1")
        self.assertEqual(data['label'], "new-test1-label")

        # test cases fro basic auth user
        data = self.client.get_app_basic_auth_users_in_group(ug_id1)
        self.assertEqual(len(data), 0)

        data = self.client.get_app_basic_auth_users_in_group(ug_id1, app_id)
        self.assertEqual(len(data), 0)

        user1 = self.client.new_app_basic_auth_user(ug_id1, "user1", "ORTest&123")
        user2 = self.client.new_app_basic_auth_user(ug_id1, "user2", "ORTest&123", app_id=app_id)
        data = self.client.get_app_basic_auth_users_in_group(ug_id1)
        self.assertEqual(len(data), 2)

        ok = self.client.del_app_basic_auth_user(user2, ug_id1)
        self.assertEqual(ok, True)

        data = self.client.get_app_basic_auth_users_in_group(ug_id1)
        self.assertEqual(len(data), 1)

        ok = self.client.put_app_basic_auth_user(user1, ug_id1, "new-user1", "ORTest&123")
        self.assertEqual(ok, True)

        ok = self.client.put_app_basic_auth_user(user1, ug_id1, "new-user1", "ORTest&123", app_id=app_id)
        self.assertEqual(ok, True)

        data = self.client.get_app_basic_auth_user(user1, ug_id1)
        self.assertEqual(data['username'], "new-user1")
        self.assertEqual(data['password'], None)

        data = self.client.get_app_basic_auth_user(user1, ug_id1, app_id)
        self.assertEqual(data['username'], "new-user1")
        self.assertEqual(data['password'], None)

    def test_basic_auth_name(self):
        app_id = self.app_id

        data = self.client.get_all_app_basic_auth_user_groups()
        self.assertEqual(len(data), 0)

        group_name = "test1"
        ug_id1 = self.client.new_app_basic_auth_user_group(group_name)

        conseq = {
            'enable-basic-authentication': {'group_name': group_name}
        }
        rule_id = self.client.new_rule(conseq=conseq)
        self.assertIs(type(rule_id), int)
        self.assertGreater(rule_id, 0)

        data = self.client.get_rule(rule_id)
        self.assertEqual(data.get('comment', None), None)
        self.assertEqual(data['actions'][0]['enable-basic-authentication']['app_auth_id'], ug_id1)



#!/usr/bin/python
# -*- coding: UTF-8 -*-
import os
import configparser

try:
    import py_client
except ImportError:
    print("Error: cannot import py_client.py")
    exit(1)

GLOBAL_VARS = []
HTML_FILES = {
    'html/50x.html': """
<!DOCTYPE html>
<html>
<head>
<title>Error</title>
<style>
    body {
        width: 35em;
        margin: 0 auto;
        font-family: Tahoma, Verdana, Arial, sans-serif;
    }
</style>
</head>
<body>
<h1>An error occurred.</h1>
<p>Sorry, the page you are looking for is currently unavailable.<br/>
Please try again later.</p>
<p>If you are the system administrator of this resource then you should check
the <a href="http://nginx.org/r/error_log">error log</a> for details.</p>
<p><em>Faithfully yours, OpenResty.</em></p>
</body>
</html>
"""
}
CLUSTER_GROUPS = []


def add_global_ngx_conf(client):
    config = {
        'ignore_invalid_headers': True,
        'ssl_session_ttl': 5,
        'ssl_session_ttl_unit': 'min',
        'ssl_protocols': [
            'SSLv3',
            'TLSv1',
            'TLSv1.1',
            'TLSv1.2'],
        'ssl_ciphers': 'EECDH+CHACHA20:EECDH+CHACHA20-draft:EECDH+AES128'
                       ':RSA+AES128:EECDH+AES256:RSA+AES256:EECDH+3DES'
                       ':RSA+3DES:!MD5',
        'ssl_prefer_server_ciphers': True}

    ok = client.set_global_ngx_config(config)
    if not ok:
        print('ERROR: set global nginx configures failed.')


def add_upstream_ngx_conf(client, conf_obj):
    if not conf_obj.has_section('upstream'):
        conf_obj.add_section('upstream')

    upstreams = client.get_all_global_upstreams()
    upstream_names = []
    for upstream in upstreams:
        upstream_names.append(upstream.get('name'))
        conf_obj.set('upstream', str(upstream.get('name')),
                     str(upstream.get('id')))


def add_variables_ngx_conf(client, conf_obj):
    if not conf_obj.has_section('variable'):
        conf_obj.add_section('variable')

    variables = client.get_all_global_vars()
    variable_names = {}

    for variable in variables:
        variable_names[variable.get('name')] = variable.get('id')

        conf_obj.set('variable',
                     str(variable.get('name')),
                     str(variable.get('id')))

    for var in GLOBAL_VARS:
        if var.get('name') not in variable_names:
            print('adding global variables ...')
            var_id = client.new_global_var(**var)
            conf_obj.set('variable', str(var.get('name')), str(var_id))


def add_actions_ngx_conf(client, conf_obj):
    if not conf_obj.has_section('action'):
        conf_obj.add_section('action')

    actions = client.get_all_global_actions()
    action_names = []

    for action in actions:
        action_names.append(action.get('name'))
        conf_obj.set('action', str(action.get('name')), str(action.get('id')))

    if 'default_set_proxy_header' not in action_names:
        print('adding global action default_set_proxy_header...')
        action_name = 'default_set_proxy_header'
        conseq = [
            {'set-proxy-header': {
                'header': 'X-Real-Scheme', 'el_var': 'scheme'}},
            {'set-proxy-header': {'header': 'Host', 'el_var': 'host'}},
            {'set-proxy-header': {
                'header': 'X-Real-IP', 'el_var': 'client-addr'}},
            {'append-proxy-header-value': {'header': 'X-Forwarded-For',
                                           'el_var': 'client-addr'}}
        ]
        action_id = \
            client.new_global_action(name=action_name, conseq=conseq)
        conf_obj.set('action', 'default_set_proxy_header', str(action_id))


def add_static_file_ngx_conf(client, conf_obj):
    static_files = client.get_all_static_files()
    static_files_labels = {}

    for static_file in static_files:
        label = static_file.get('label')
        if label:
            static_files_labels[label] = static_file.get('id')

    if not conf_obj.has_section('favicon'):
        conf_obj.add_section('favicon')

    if not conf_obj.has_section('html'):
        conf_obj.add_section('html')

    for html_label, html_content in HTML_FILES.items():
        if html_label not in static_files_labels:
            print('adding html static files ...')
            file_id = client.upload_static_file(
                filename=html_label,content=html_content, label=html_label)
            conf_obj.set('html', str(html_label), str(file_id))
        else:
            conf_obj.set('html', str(html_label),
                         str(static_files_labels[html_label]))


def add_ssl_ngx_conf(client, conf_obj):
    if not conf_obj.has_section('ssl_cert'):
        conf_obj.add_section('ssl_cert')

    global_certs = client.get_all_global_cert_key()
    print(global_certs)
    cert_labels = {}

    for cert in global_certs:
        if cert.get('label'):
            cert_labels[cert.get('label')] = cert.get('id')


def add_cluster_ngx_conf(client, conf_obj):
    if not conf_obj.has_section('cluster'):
        conf_obj.add_section('cluster')

    all_cluster_groups = client.get_all_cluster_groups()
    cluster_names = {}
    for cluster_group in all_cluster_groups:
        if cluster_group.get('name') and cluster_group.get('id'):
            cluster_names[cluster_group.get('name')] \
                = cluster_group.get('id')

            conf_obj.set('cluster',
                         str(cluster_group.get('name')),
                         str(cluster_group.get('id')))

    for cluster_name in CLUSTER_GROUPS:
        if cluster_name not in cluster_names:
            print('adding cluster group ...')
            cluster_id = client.new_cluster_group(cluster_name)
            conf_obj.set('cluster', str(cluster_name), str(cluster_id))


def main():
    client = py_client.get_client()
    if not client:
        exit(1)

    conf_obj = configparser.ConfigParser()

    print('adding global nginx configures...')
    add_global_ngx_conf(client)

    cwd = os.getcwd()
    global_config_file_path = os.path.join(cwd, "global.ini")

    # read old global configures
    if os.path.isfile(global_config_file_path):
        conf_obj.read(global_config_file_path, encoding="utf-8")

    print('adding upstream nginx configures...')
    add_upstream_ngx_conf(client, conf_obj)

    print('adding variables nginx configures...')
    add_variables_ngx_conf(client, conf_obj)

    print('adding actions nginx configures...')
    add_actions_ngx_conf(client, conf_obj)

    print('adding static file nginx configures...')
    add_static_file_ngx_conf(client, conf_obj)

    print('adding ssl nginx configures...')
    add_ssl_ngx_conf(client, conf_obj)

    print('adding cluster nginx configures...')
    add_cluster_ngx_conf(client, conf_obj)

    print('writing config.ini...: ' + global_config_file_path)
    with open(global_config_file_path, 'w') as cf_file:
        conf_obj.write(cf_file)

    print('all done')


if __name__ == '__main__':
    main()




# -*- coding: utf-8 -*-

from .edge2client import Edge2Client
import os

version = {}
with open(os.path.join(os.path.dirname(__file__), "version.py")) as fp:
    exec(fp.read(), version)

__version__ = version['__version__']
__title__ = 'edge2client'
__copyright__ = 'Copyright 2020 OpenResty Inc.'

__all__ = [
    'Edge2Client', '__version__'
]



# -*- coding: utf-8 -*-

from .constants import OFF, \
    waf_rule_sets_name2id, \
    valid_waf_thresholds, \
    cross_req_waf_threshold_scores, \
    waf_threshold_scores

def process_rule_cond(condition):
    if not condition:
        raise Exception('No condition field specified')

    if not isinstance(condition, list):
        raise Exception(
            'Bad condition field value type: ' + str(type(condition)))

    i = 0
    cond_specs = []

    for cond in condition:
        i += 1
        operator = cond.get('op', 'eq')
        var = cond.get('var')
        global_var = cond.get('global_var')
        user_var = cond.get('user_var')
        if not var and not global_var and not user_var:
            raise Exception(
                'No var or global_var or user_var '
                'specified in the {}-th condition'.format(str(i)))

        var_arg = None
        if isinstance(var, list):
            if len(var) != 2:
                raise Exception(
                    'variable spec must take 2 '
                    'elements in the {}-th condition'.format(str(i)))
            var, var_arg = var

        val = cond.get('val')
        vals = cond.get('vals')
        if val is not None:
            if vals is not None:
                raise Exception(
                    'The val and vals fields cannot be '
                    'specified at the same time')
            vals = [val]
            del val

        val_specs = []
        if operator not in ('is-empty', '!is-empty'):
            for val in vals:
                val_type = 'str'
                if isinstance(val, list):
                    if len(val) != 2:
                        raise Exception(
                            'variable spec must take 2 elements in the '
                            '{}-th condition'.format(str(i)))
                    val, val_type = val
                val_specs.append({'type': val_type, 'val': val})

        if var:
            variable = {'name': var}
            if var_arg:
                variable['args'] = var_arg
        elif global_var:
            variable = {'global_var': global_var}
        else:
            variable = {'user_var': user_var}

        cond_spec = {
            'variable': variable,
            'operator': {'name': operator},
        }

        if val_specs:
            cond_spec['values'] = val_specs

        caseless = cond.get('caseless')
        if caseless:
            cond_spec['caseless'] = caseless

        cond_specs.append(cond_spec)

    return cond_specs


def process_conseq(conseq):
    if not conseq:
        raise Exception('No conseq field specified')
    if isinstance(conseq, dict):
        conseq = [conseq]
    elif not isinstance(conseq, list):
        raise Exception('Bad conseq field value type: '
                        + str(type(conseq)))

    conseq_specs = []

    for item in conseq:
        for action_name, action_spec in item.items():
            if action_name == "global_action":
                conseq_specs.append({'global_action_id': action_spec})
                continue

            conseq_spec = {'type': action_name}
            if action_spec:
                conseq_spec[action_name] = action_spec
            conseq_specs.append(conseq_spec)

    return conseq_specs


def process_waf(waf):
    rule_sets = waf.get('rule_sets', [])

    rule_set_ids = dict()
    for rule_set_id in rule_sets:
        rule_set_ids[rule_set_id] = 1

    for (name, _) in waf_rule_sets_name2id.items():
        if waf.get(name, None):
            rule_set_id = waf_rule_sets_name2id[name]
            if rule_set_id and rule_set_ids.get(rule_set_id, None) is None:
                rule_sets.append(rule_set_id)

    action = waf.get('action', None)
    threshold = waf.get('threshold', 'medium')
    clearance = waf.get('clearance', 60)
    redirect_url = waf.get('redirect_url', None)
    page_template_id = waf.get('page_template_id', None)
    page_template_status_code = waf.get('page_template_status_code', 200)
    cross_requests = waf.get('cross_requests', True)
    sensitivity = waf.get('sensitivity', None)
    score = waf.get('score', None)
    rule_sets_threshold = waf.get('rule_sets_threshold', None)

    # sensitivity has a higher priority than threshold
    if sensitivity == "high":
        threshold = "low"
    elif sensitivity == "low":
        threshold = "high"
    elif sensitivity:
        threshold = sensitivity

    if sensitivity and not valid_waf_thresholds.get(sensitivity, None):
        raise Exception('Unsupported sensitivity level: {}'.format(sensitivity))

    if threshold and not valid_waf_thresholds.get(threshold, None):
        raise Exception('Unsupported threshold level: {}'.format(threshold))

    if not action:
        raise Exception('No action field specified')

    if action == '403 Forbidden':
        action = 'block'

    if cross_requests == True:
        threshold_score = cross_req_waf_threshold_scores.get(threshold, None)
    else:
        threshold_score = waf_threshold_scores.get(threshold, None)

    if threshold_score is None:
        if score:
            if isinstance(score, int) == False or score <= 0:
                raise Exception('No value of score field')
            # use the value passed by the user
            threshold_score = score
        else:
            raise Exception('No value of threshold field')

    body = {
        'rule_sets': rule_sets,
        'action': action,
        'threshold_score': threshold_score,
        'cross_requests': cross_requests,
    }

    if cross_requests == False:
        if rule_sets_threshold:
            body['rule_sets_threshold'] = rule_sets_threshold
        else:
            body['rule_sets_threshold'] = [-1 for _ in range(len(rule_sets))]

    if action == 'edge-captcha' and clearance:
        body['clearance_time'] = clearance

    if action == 'redirect' and redirect_url:
        body['redirect_url'] = redirect_url

    if action == 'page-template' and page_template_id:
        body['page_template_id'] = page_template_id
        body['page_template_status_code'] = page_template_status_code

    return body


def process_proxy(proxy):
    upstream_el_code = proxy.get('upstream_el_code', '')
    upstreams = proxy.get('upstreams', None) or proxy.get('upstream', [])
    if not upstreams and not upstream_el_code:
        raise Exception('No upstreams specified')
    if upstreams and upstream_el_code:
        raise Exception('upstreams and upstream_el_code are conflicting')

    backup_upstreams = proxy.get('backup_upstreams', None) or proxy.get('backup_upstream', {})

    timeout = proxy.get('timeout', 3)
    connect_timeout = proxy.get('connect_timeout', timeout)
    read_timeout = proxy.get('read_timeout', timeout)
    send_timeout = proxy.get('send_timeout', timeout)
    retries = proxy.get('retries', 5)
    retry_condition = proxy.get('retry_condition',
                                ["error",
                                 "timeout",
                                 "invalid_header",
                                 "http_500",
                                 "http_502",
                                 "http_504",
                                 "http_404"])
    balancer = proxy.get('balancer', {})
    balancer_algorithm = proxy.get('balancer_algorithm', None) or balancer.get('algorithm', "roundrobin")
    balancer_vars = proxy.get('balancer_vars', None) or balancer.get('variables', [])
    multi_tier = proxy.get('multi_tier', None) or proxy.get('layer_policy', None)
    sticky = proxy.get('sticky', {'level': 'upstream','enable': False,
                                'ttl': 0,'key': 'Edge-Sticky','mode': 'none'})


    if balancer_algorithm != 'roundrobin' and balancer_vars == []:
        raise Exception('No balancer vars specified')

    up_specs = []
    for upstream in upstreams:
        up_id = upstream.get('upstream', None)
        k8s_up_id = upstream.get('k8s_upstream', None)
        global_up_id = upstream.get('global_upstream', None)
        global_k8s_up_id = upstream.get('global_k8s_upstream', None)
        weight = upstream.get('weight', 1)

        up_spec = {
            'cluster': up_id,
            'k8s_cluster': k8s_up_id,
            'global_cluster': global_up_id,
            'global_k8s_cluster': global_k8s_up_id,
            'weight': weight,
        }
        up_specs.append(up_spec)

    bup_specs = []
    for upstream in backup_upstreams:
        up_id = upstream.get('upstream', None)
        k8s_up_id = upstream.get('k8s_upstream', None)
        global_up_id = upstream.get('global_upstream', None)
        global_k8s_up_id = upstream.get('global_k8s_upstream', None)
        weight = upstream.get('weight', 1)

        bup_spec = {
            'cluster': up_id,
            'k8s_cluster': k8s_up_id,
            'global_cluster': global_up_id,
            'global_k8s_cluster': global_k8s_up_id,
            'weight': weight,
        }
        bup_specs.append(bup_spec)

    data = {
        'upstream': up_specs,
        'upstream_el_code': upstream_el_code,
        'connect_timeout': connect_timeout,
        'send_timeout': send_timeout,
        'read_timeout': read_timeout,
        'retries': retries,
        'retry_condition': retry_condition,
        'balancer': {
            'algorithm': balancer_algorithm,
            'variables': balancer_vars
        },
        'sticky': sticky,
    }

    if bup_specs:
        data['backup_upstream'] = bup_specs

    if multi_tier:
        data['layer_policy'] = multi_tier

    return data


def process_proxy_patch(proxy):
    upstream_el_code = proxy.get('upstream_el_code', '')
    upstreams = proxy.get('upstreams', None)
    if not upstreams and not upstream_el_code:
        raise Exception('No upstreams specified')
    if upstreams and upstream_el_code:
        raise Exception('upstreams and upstream_el_code are conflicting')

    backup_upstreams = proxy.get('backup_upstreams')
    connect_timeout = proxy.get('connect_timeout')
    read_timeout = proxy.get('read_timeout')
    send_timeout = proxy.get('send_timeout')
    retries = proxy.get('retries')
    retry_condition = proxy.get('retry_condition')
    balancer_algorithm = proxy.get('balancer_algorithm')
    balancer_vars = proxy.get('balancer_vars')
    multi_tier = proxy.get('multi_tier')

    if balancer_algorithm:
        if balancer_algorithm != 'roundrobin' and not balancer_vars:
            raise Exception('No balancer vars specified')

    up_specs = []
    for upstream in upstreams:
        up_id = upstream.get('upstream')
        global_up_id = upstream.get('global_upstream')
        weight = upstream.get('weight')

        # weight is zero means remove it from proxy_rule.
        if weight == 0:
            continue

        up_spec = {
            'cluster': up_id,
            'global_cluster': global_up_id,
            'weight': weight,
        }
        up_specs.append(up_spec)

    bup_specs = []
    if backup_upstreams:
        for upstream in backup_upstreams:
            up_id = upstream.get('upstream')
            weight = upstream.get('weight')

            # weight is zero means remove it from proxy_rule.
            if weight == 0:
                continue
            bup_spec = {
                'cluster': up_id,
                'weight': weight,
            }
            bup_specs.append(bup_spec)

    data = {'upstream': up_specs}

    if connect_timeout:
        data['connect_timeout'] = connect_timeout
    if send_timeout:
        data['send_timeout'] = send_timeout
    if read_timeout:
        data['read_timeout'] = read_timeout
    if retries:
        data['retries'] = retries
    if retry_condition:
        data['retry_condition'] = retry_condition
    if balancer_algorithm:
        data['balancer'] = {'algorithm': balancer_algorithm}

        if balancer_vars:
            data['balancer']['variables'] = balancer_vars

    if bup_specs:
        data['backup_upstream'] = bup_specs

    if multi_tier:
        data['layer_policy'] = multi_tier

    if upstream_el_code is not None:
        data['upstream_el_code'] = upstream_el_code

    return data

def is_valid_unit(unit):
    if unit == 's' or unit == 'min' or unit == 'hour' or unit == 'day':
        return True

    return False

def is_valid_cache_status(status):
    if status == 200 or status == 301 or status == 302:
        return True

    return False

def process_cache(cache):
    cache_key = cache.get('cache_key', None)
    default_ttls = cache.get('default_ttls', None)
    browser_ttl  = cache.get('browser_ttl', None)
    browser_ttl_unit = cache.get('browser_ttl_unit', 'min')
    enable_global = cache.get('enable_global', False)
    enforce_cache = cache.get('enforce_cache', False)
    cluster_hash = cache.get('cluster_hash', False)
    disable_convert_head = cache.get('disable_convert_head', True)

    if not cache_key:
        raise Exception('Missing value for the required field: cache_key')
    if not isinstance(cache_key, list):
        raise Exception('Bad cache_key field value type: ' + str(type(cache_key)))

    if default_ttls and not isinstance(default_ttls, list):
        raise Exception('Bad default_ttls field value type: ' + str(type(default_ttls)))

    cache_key_specs = []
    for item in cache_key:
        if isinstance(item, str):
            cache_key_specs.append({'name': item})
        else:
            name = item.get('name')
            args = item.get('args', None)

            cache_key_specs.append({'name': name, 'args': args})

    data = {'cache_key': cache_key_specs}

    if enforce_cache:
        # False or empty list
        if not default_ttls:
            raise Exception('Missing default_ttls for enforce_cache')

    data['enforce_cache'] = enforce_cache

    if default_ttls == OFF:
        data['default_ttls'] = None

    elif default_ttls:
        data['default_ttls'] = list()
        for item in default_ttls:
            ttl = item.get('ttl', 1)
            ttl_unit = item.get('ttl_unit', 'min')
            if not is_valid_unit(ttl_unit):
                raise Exception('Bad default_ttls unit: ' + str(ttl_unit))

            status = item.get('status', 200)
            if not is_valid_cache_status(status):
                raise Exception('Bad default_ttls status: ' + str(status))

            data['default_ttls'].append({'ttl': ttl, 'ttl_unit': ttl_unit, 'status': status})

    if browser_ttl == OFF:
        data['browser_ttl'] = None
        data['browser_ttl_unit'] = None

    elif browser_ttl is not None:
        data['browser_ttl'] = browser_ttl
        data['browser_ttl_unit'] = browser_ttl_unit
        if not is_valid_unit(browser_ttl_unit):
            raise Exception('Bad browser_ttl unit: ' + str(browser_ttl_unit))

    data['cluster_hash'] = cluster_hash
    data['enable_global'] = enable_global
    data['disable_convert_head'] = disable_convert_head

    return data


def process_content(content):
    file = content.get('file', None)
    favicon = content.get('favicon', None)
    empty_gif = content.get('empty_gif', None)

    if favicon:
        return {'file': favicon, 'ignore_uri_prefix_type': None, 'ignore_uri_prefix_value': None, 'type': 'file'}
    if empty_gif:
        return {'empty_gif': True}
    if file:
        return content

    raise Exception(
        'Missing value for the required field: favicon or empty_gif')

def proxy_upstream_transform(upstream):
    if upstream.get('cluster') is not None:
        upstream_id = upstream.get('cluster')
        upstream['upstream'] = upstream_id
        del upstream['cluster']
    if upstream.get('k8s_cluster') is not None:
        upstream_id = upstream.get('k8s_cluster')
        upstream['k8s_upstream'] = upstream_id
        del upstream['k8s_cluster']
    if upstream.get('global_cluster') is not None:
        upstream_id = upstream.get('global_cluster')
        upstream['global_upstream'] = upstream_id
        del upstream['global_cluster']
    if upstream.get('global_k8s_cluster') is not None:
        upstream_id = upstream.get('global_k8s_cluster')
        upstream['global_k8s_upstream'] = upstream_id
        del upstream['global_k8s_cluster']

def get_first_ip(ips_str):
    ips = ips_str.split(' ')
    return ips[0]



#!/usr/bin/python
# -*- coding: UTF-8 -*-
from os import getcwd, path, system, remove
import sys

IMPORT_CODE = 'import py_client\nclient = py_client.get_client()\n'


def main(app_file):
    if path.isfile(app_file):
        with open(app_file) as file_handler:
            content = file_handler.read()
            name = path.basename(app_file)
            tmp_file = path.join(path.dirname(__file__), name)
            with open(tmp_file, 'w') as file_writer:
                file_writer.write(IMPORT_CODE)
                file_writer.write(content)
            print('-------------- init ------------\n')
            print('copy <{}> to <{}>'.format(app_file, tmp_file))
            system(sys.executable + ' ' + tmp_file)
            print('created app with ' + name)
            remove(tmp_file)
    else:
        print(app_file + ' is not a file')


if __name__ == '__main__':
    if len(sys.argv) > 1 and sys.argv[1]:
        APP_FILE = sys.argv[1]
        main(APP_FILE)
    else:
        print('app_file required')



import os

from .utils import error, warn, info, fix_line, is_valid_host, \
    is_valid_ipv4_address, is_valid_ipv6_address
from .read_config import read_yaml_config, write_yaml_config, add_before_comment

def check_k8s_clusters(clusters, filename):
    if not isinstance(clusters, list):
        error(f"Unsupported K8s clusters file format, file: {filename}")

    names = dict()
    for cluster in clusters:
        if not isinstance(cluster, dict):
            error(f"Each K8s cluster must be a dictionary, file: {filename}, line: {fix_line(cluster.lc.line)}")

        required_fields = ['name', 'host', 'port', 'token']
        for field in required_fields:
            if field not in cluster:
                error(f"Missing required field '{field}' in K8s cluster, file: {filename}, line: {fix_line(cluster.lc.line)}")

        if cluster['name'] in names:
            error(f"Duplicate name in K8s cluster {cluster['name']}, file: {filename}, line: {fix_line(cluster.lc.line)}")

        names[cluster['name']] = True

        if not is_valid_host(cluster['host']):
            error(f"Invalid host in K8s cluster {cluster['name']}: {cluster['host']}, file: {filename}, line: {fix_line(cluster.lc.line)}")

        if not isinstance(cluster['port'], int) or not (1 <= cluster['port'] <= 65535):
            error(f"Invalid port in K8s cluster {cluster['name']}: {cluster['port']}, file: {filename}, line: {fix_line(cluster.lc.line)}")

        if not isinstance(cluster['token'], str) or not cluster['token']:
            error(f"Invalid token in K8s cluster {cluster['name']}, file: {filename}, line: {fix_line(cluster.lc.line)}")

        if 'ssl_verify' in cluster and not isinstance(cluster['ssl_verify'], bool):
            error(f"Invalid ssl_verify in K8s cluster {cluster['name']}, file: {filename}, line: {fix_line(cluster.lc.line)}")

        for timeout in ['connect_timeout', 'read_timeout', 'send_timeout']:
            if timeout in cluster and not isinstance(cluster[timeout], int):
                error(f"Invalid {timeout} in K8s cluster {cluster['name']}, file: {filename}, line: {fix_line(cluster.lc.line)}")

    return True

def is_change_k8s_cluster(new_cluster, old_cluster):
    fields_to_compare = [
        'host',
        'port',
        'domain',
        'ssl_verify',
        'connect_timeout',
        'read_timeout',
        'send_timeout',
        'token'
    ]
    return any(new_cluster.get(field) != old_cluster.get(field) for field in fields_to_compare)

def process_k8s_clusters(ctx):
    client = ctx['client']
    configs_path = ctx['configs_path']
    location = ctx.get('location', None)

    if not location:
        location = "k8s"

    configs = read_yaml_config(configs_path, location)
    if configs is None:
        return

    info("Checking if K8s clusters are valid")
    for filename, clusters in configs.items():
        check_k8s_clusters(clusters, filename)

    old_clusters = client.get_all_global_k8s()

    old_clusters_dict = {cluster['name']: cluster for cluster in old_clusters}

    info("Checking if K8s clusters have changed")
    new_clusters_dict = {}
    sorted_configs = sorted(configs.keys())

    for filename in sorted_configs:
        clusters = configs[filename]
        for cluster in clusters:
            new_clusters_dict[cluster['name']] = True

            host = None
            domain = None
            if is_valid_ipv4_address(cluster['host']) or is_valid_ipv6_address(cluster['host']):
                host = cluster['host']
            else:
                domain = cluster['host']

            if cluster['name'] in old_clusters_dict:
                old_cluster = old_clusters_dict[cluster['name']]
                if is_change_k8s_cluster(cluster, old_cluster):
                    try:
                        info(f"Updating K8s cluster \"{cluster['name']}\"")
                        client.put_global_k8s(
                            k8s_id=old_cluster['id'],
                            name=cluster['name'],
                            host=host,
                            domain=domain,
                            port=cluster['port'],
                            ssl_verify=cluster.get('ssl_verify', False),
                            connect_timeout=cluster.get('connect_timeout', 15),
                            read_timeout=cluster.get('read_timeout', 15),
                            send_timeout=cluster.get('send_timeout', 15),
                            token=cluster['token']
                        )
                    except Exception as e:
                        error(f"Failed to update K8s cluster, file: {filename}, line: {fix_line(cluster.lc.line)}", e)
            else:
                try:
                    info(f"Adding K8s cluster \"{cluster['name']}\"")
                    client.new_global_k8s(
                        name=cluster['name'],
                        host=host,
                        domain=domain,
                        port=cluster['port'],
                        ssl_verify=cluster.get('ssl_verify', False),
                        connect_timeout=cluster.get('connect_timeout', 30),
                        read_timeout=cluster.get('read_timeout', 30),
                        send_timeout=cluster.get('send_timeout', 30),
                        token=cluster['token']
                    )
                except Exception as e:
                    error(f"Failed to add K8s cluster, file: {filename}, line: {fix_line(cluster.lc.line)}", e)

    # for cluster_name, cluster in old_clusters_dict.items():
    #     if cluster_name not in new_clusters_dict:
    #         try:
    #             info(f"Removing K8s Endpoint in K8s cluster \"{cluster_name}\"")
    #             client.del_k8s_endpoints(cluster['id'])
    #             info(f"Removing K8s cluster \"{cluster_name}\"")
    #             client.del_global_k8s(cluster['id'])
    #         except Exception as e:
    #             error(f"Failed to remove K8s cluster, cluster id: {cluster['id']}", e)

def cleanup_k8s_clusters(ctx):
    pass
    # client = ctx['client']

    # clusters = client.get_all_global_k8s()

    # for cluster in clusters:
    #     try:
    #         info(f"Removing K8s Endpoint in K8s cluster \"{cluster['name']}\"")
    #         client.del_k8s_endpoints(cluster['id'])
    #         info(f"Removing K8s cluster \"{cluster['name']}\"")
    #         client.del_global_k8s(cluster['id'])
    #     except Exception as e:
    #         error(f"Failed to remove K8s cluster, cluster id: {cluster['id']}", e)

def export_k8s_clusters(ctx):
    client = ctx['client']
    configs_path = ctx['export_to_path']

    clusters = client.get_all_global_k8s()

    if not clusters:
        info(f"No K8s clusters found")
        return

    formatted_clusters = []
    for cluster in clusters:
        host = cluster.get('host', None)
        if host is None:
            host = cluster.get('domain', None)

        token = ''
        if ctx['export_fake_info']:
            token = "********"

        formatted_cluster = {
            'name': cluster['name'],
            'host': host,
            'port': cluster['port'],
            'ssl_verify': cluster.get('ssl_verify', False),
            'connect_timeout': cluster.get('connect_timeout', 15),
            'read_timeout': cluster.get('read_timeout', 15),
            'send_timeout': cluster.get('send_timeout', 15),
            'token': token,
        }

        formatted_cluster = add_before_comment(formatted_cluster, 'token',
                                                "token: We don't export actual token for security reasons")

        formatted_clusters.append(formatted_cluster)

    export_path = os.path.join(configs_path, "k8s")

    try:
        write_yaml_config(export_path, "k8s.yaml", formatted_clusters)
        info(f"K8s clusters exported successfully to k8s/k8s.yaml")
    except Exception as e:
        error(f"Failed to export K8s clusters to k8s/k8s.yaml", e)



# -*- coding: utf-8 -*-
# import io
import sdk_test


class TestApp(sdk_test.TestSdk):
    def test_app(self):
        app_id = self.app_id

        ok = self.client.put_app(app_id=app_id, domains=[self.apex],
                                 label=self.apex, cluster_groups=[1],
                                 offline=False, http_ports=[80,81],
                                 https_ports=[443,444], is_default=True)
        self.assertTrue(ok)

        self.client.new_release()

        ok = self.client.append_app_domain(app_id=app_id,
                                           domain='foo.' + self.apex)
        self.assertTrue(ok)

        data = self.client.get_app()

        self.assertEqual(data['name'], self.apex)
        self.assertEqual(data['partitions'], [1])
        self.assertEqual(data['domains'][0]['domain'], self.apex)
        self.assertEqual(data['domains'][1]['domain'], 'foo.' + self.apex)
        self.assertFalse(data['domains'][1]['is_wildcard'])
        self.assertEqual(data['is_default'], True)

        limiter = {
            'max_uri_args': 120,
            'max_post_args': 130,
            'max_req_headers': 140
        }
        self.client.put_app_config(app_id=app_id, limiter=limiter, enable_websocket=True)
        self.client.new_release()
        config = self.client.get_app_config(app_id=app_id)
        self.assertEqual(config['limiter']['max_uri_args'], 120)
        self.assertEqual(config['limiter']['max_post_args'], 130)
        self.assertEqual(config['limiter']['max_req_headers'], 140)
        self.assertTrue(config['enable_websocket'])

        self.client.put_app_config(app_id=app_id, client_max_body_size=1024, client_max_body_size_unit='m')
        self.client.new_release()
        config = self.client.get_app_config(app_id=app_id)
        self.assertEqual(config['limiter']['max_uri_args'], 120)
        self.assertEqual(config['limiter']['max_post_args'], 130)
        self.assertEqual(config['limiter']['max_req_headers'], 140)
        self.assertEqual(config['client_max_body_size'], 1024)
        self.assertEqual(config['client_max_body_size_unit'], 'm')
        self.assertTrue(config['enable_websocket'])

        self.client.put_app_config(app_id=app_id, client_max_body_size=None)
        self.client.new_release()
        config = self.client.get_app_config(app_id=app_id)
        self.assertEqual(config['limiter']['max_uri_args'], 120)
        self.assertEqual(config['limiter']['max_post_args'], 130)
        self.assertEqual(config['limiter']['max_req_headers'], 140)
        self.assertEqual(config.get('client_max_body_size', None), None)
        self.assertEqual(config.get('client_max_body_size_unit', None), None)
        self.assertTrue(config['enable_websocket'])

    def test_search_app_by_domain_and_get_all_apps(self):
        app_id = self.app_id

        data = self.client.search_app(app_domain=self.apex)
        self.assertEqual(len(data), 1)
        self.assertEqual(data[0]['name'], self.apex)
        self.assertEqual(data[0]['id'], app_id)

        new_app_id = self.client.new_app(domains=['prefix' + self.apex],
                                         label='prefix' + self.apex)
        self.assertIs(type(new_app_id), int)
        self.assertGreater(new_app_id, 0)

        data = self.client.search_app(app_domain=self.apex)
        self.assertEqual(len(data), 2)
        self.assertEqual(data[0]['domains'][0]['domain'], self.apex)
        self.assertEqual(data[0]['id'], app_id)
        self.assertEqual(data[1]['domains'][0]['domain'], 'prefix' + self.apex)
        self.assertEqual(data[1]['id'], new_app_id)

        apps = self.client.get_all_apps()
        for app in apps:
            self.assertIs(type(app), int)
            self.assertGreater(app, 0)

        apps = self.client.get_all_apps(detail=True)
        self.assertEqual(len(apps), 2)
        self.assertEqual(apps[app_id]['label'], self.apex)
        self.assertEqual(apps[app_id]['domains'][0]['domain'], self.apex)
        self.assertEqual(apps[new_app_id]['label'], 'prefix' + self.apex)
        self.assertEqual(apps[new_app_id]['domains'][0]['domain'],
                         'prefix' + self.apex)

        ok = self.client.del_app(new_app_id)
        self.assertTrue(ok)

        self.client.use_app(app_id)




# -*- coding: utf-8 -*-
# import io
import sdk_test

class TestGlobalDymetrics(sdk_test.TestSdk):
    def test_global_dymetrics(self):
        sql = 'select status, count(*) from reqs group by status'
        dy_id = self.client.new_global_dymetrics(name='test', note='abc', interval=60, sql=sql)
        self.assertIs(type(dy_id), int)

        data = self.client.get_global_dymetrics(dy_id)
        self.assertEqual(data['name'], 'test')

        sql = 'select status, count(*) from reqs group by status'
        dy_id2 = self.client.new_global_dymetrics(name='test2', interval=120, sql=sql)
        self.assertIs(type(dy_id2), int)

        ok = self.client.put_global_dymetrics(dy_id, name='testput')
        self.assertTrue(ok)

        data = self.client.get_global_dymetrics(dy_id)
        self.assertEqual(data['name'], 'testput')

        data = self.client.get_global_dymetrics_data(dy_id)
        self.assertEqual(data, [])

        data = self.client.get_all_global_dymetrics()
        self.assertGreaterEqual(len(data), 2)

        ok = self.client.del_global_dymetrics(dy_id)
        self.assertTrue(ok)

        ok = self.client.del_global_dymetrics(dy_id2)
        self.assertTrue(ok)



# -*- coding: utf-8 -*-
import os
import string
import random
import time
import subprocess
import sdk_test

CURPATH = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))

class TestErrorLogs(sdk_test.TestSdk):
    app_id = None

    def test_001_init(self):
        init_script = os.path.join(CURPATH, 'tests', 'data', 'init-error-logs.sh')

        app_id = self.app_id
        TestErrorLogs.app_id = app_id

        command = "/usr/bin/bash {} {}".format(init_script, app_id)
        result = subprocess.run(command, shell=True, capture_output=True, text=True)
        print(result.stdout)
        if result.returncode != 0:
            print(result.stderr)
        self.assertEqual(result.returncode, 0)

    def test_002_node_error_logs(self):
        data, count = self.client.get_node_error_logs()
        self.assertEqual(len(data), count)
        self.assertTrue(count >= 0)

        start_time = 1728345983
        data, count = self.client.get_node_error_logs(start_time=start_time)
        self.assertEqual(len(data), 10)
        self.assertTrue(count == 20)

        end_time = 1728346598
        data, count = self.client.get_node_error_logs(start_time=start_time, end_time=end_time)
        self.assertEqual(len(data), 10)
        self.assertTrue(count == 11)

        data, count = self.client.get_node_error_logs(start_time=start_time, req_id="000028800bac65d7fed03afd")
        self.assertEqual(len(data), 2)
        self.assertTrue(count == 2)

        data, count = self.client.get_node_error_logs(start_time=start_time, level="error")
        self.assertEqual(len(data), 10)
        self.assertTrue(count == 20)

        data, count = self.client.get_node_error_logs(start_time=start_time, level="warn")
        self.assertEqual(len(data), 10)
        self.assertTrue(count == 20)

        data, count = self.client.get_node_error_logs(start_time=start_time, level="crit")
        self.assertEqual(len(data), 0)
        self.assertTrue(count == 0)


    def test_003_admin_error_logs(self):
        data, count = self.client.get_admin_error_logs()
        self.assertEqual(len(data), count)
        self.assertTrue(count >= 0)

        start_time = 1528345600
        data, count = self.client.get_admin_error_logs(start_time=start_time)
        self.assertEqual(len(data), 10)
        self.assertTrue(count == 20)

        end_time = 1728345629
        data, count = self.client.get_admin_error_logs(start_time=start_time, end_time=end_time)
        self.assertEqual(len(data), 5)
        self.assertTrue(count == 5)

        data, count = self.client.get_admin_error_logs(start_time=start_time, req_id="000028800bac65d7fed03afd")
        self.assertEqual(len(data), 0)
        self.assertTrue(count == 0)

        data, count = self.client.get_admin_error_logs(start_time=start_time, level="error")
        self.assertEqual(len(data), 0)
        self.assertTrue(count == 0)

        data, count = self.client.get_admin_error_logs(start_time=start_time, level="warn")
        self.assertEqual(len(data), 10)
        self.assertTrue(count == 20)

    def test_004_log_server_error_logs(self):
        data, count = self.client.get_log_server_error_logs()
        self.assertEqual(len(data), count)
        self.assertTrue(count >= 0)

        start_time = 1528345600
        data, count = self.client.get_log_server_error_logs(start_time=start_time)
        self.assertEqual(len(data), 10)
        self.assertTrue(count == 20)

        end_time = 1728349200
        data, count = self.client.get_log_server_error_logs(start_time=start_time, end_time=end_time)
        self.assertEqual(len(data), 6)
        self.assertTrue(count == 6)

        data, count = self.client.get_log_server_error_logs(start_time=start_time, req_id="000028800bac65d7fed03afd")
        self.assertEqual(len(data), 0)
        self.assertTrue(count == 0)

        data, count = self.client.get_log_server_error_logs(start_time=start_time, level="error")
        self.assertEqual(len(data), 0)
        self.assertTrue(count == 0)

        data, count = self.client.get_log_server_error_logs(start_time=start_time, level="warn")
        self.assertEqual(len(data), 10)
        self.assertTrue(count == 20)

    def test_005_http_app_error_logs(self):
        self.app_id = TestErrorLogs.app_id
        self.client.use_app(self.app_id)

        data, count = self.client.get_http_app_error_logs()
        self.assertEqual(len(data), count)
        self.assertTrue(count >= 0)

        start_time = 1528345600
        data, count = self.client.get_http_app_error_logs(start_time=start_time)
        self.assertEqual(len(data), 10)
        self.assertTrue(count == 20)

        end_time = 1728346207
        data, count = self.client.get_http_app_error_logs(start_time=start_time, end_time=end_time)
        self.assertEqual(len(data), 5)
        self.assertTrue(count == 5)

        data, count = self.client.get_http_app_error_logs(start_time=start_time, req_id="0000270000b465d7f2f037f5")
        self.assertEqual(len(data), 1)
        self.assertTrue(count == 1)

        data, count = self.client.get_http_app_error_logs(start_time=start_time, level="crit")
        self.assertEqual(len(data), 0)
        self.assertTrue(count == 0)

        data, count = self.client.get_http_app_error_logs(start_time=start_time, level="error")
        self.assertEqual(len(data), 10)
        self.assertTrue(count == 20)

        data, count = self.client.get_http_app_error_logs(start_time=start_time, level="warn")
        self.assertEqual(len(data), 10)
        self.assertTrue(count == 20)



# -*- coding: utf-8 -*-
# import io
import sdk_test
import random
import time

class TestUpstream(sdk_test.TestSdk):
    def test_upstream_healthcheck_status(self):
        up_id = self.client.new_upstream(
            name='origin-upstream',
            servers=[
                {'domain': 'test.com', 'port': 80},
                {'ip': '172.22.31.2', 'port': 80, 'weight': 2}
            ],
            health_checker={
                'http_req_uri': '/status',
                'http_req_host': 'test.com',
                'interval': 3,
                'interval_unit': 'sec',
                'timeout': 1,
                'fall': 3,
                'rise': 2,
                'valid_statuses': [200, 302],
                'report_interval': 3,
                'report_interval_unit': 'min'
            })
        self.assertIs(type(up_id), int)
        self.assertGreater(up_id, 0)

        data = self.client.get_upstream_healthcheck_status(self.app_id, up_id)
        self.assertIs(type(data), dict)
        data = data['data']
        self.assertIs(type(data), list)
        self.assertEqual(len(data), 0)


    def test_global_upstream_healthcheck_status(self):
        now = random.randint(1, int(time.time()))
        up_id = self.client.new_global_upstream(
            name='origin-upstream' + str(now),
            servers=[
                {'domain': 'test.com', 'port': 80},
                {'ip': '172.22.31.2', 'port': 80, 'weight': 2}
            ],
            health_checker={
                'http_req_uri': '/status',
                'http_req_host': 'test.com',
                'interval': 3,
                'interval_unit': 'sec',
                'timeout': 1,
                'fall': 3,
                'rise': 2,
                'valid_statuses': [200, 302],
                'report_interval': 3,
                'report_interval_unit': 'min'
            })
        self.assertIs(type(up_id), int)
        self.assertGreater(up_id, 0)

        data = self.client.get_global_upstream_healthcheck_status(up_id)
        self.assertIs(type(data), dict)
        data = data['data']
        self.assertIs(type(data), list)
        self.assertEqual(len(data), 0)



import os

from .utils import error, warn, info, get_app_id_by_domain, line
from .read_config import read_yaml_config, write_yaml_config
from .page_rules import export_conditions, build_names_map

def delete_waf_whitelist(client, app_id):
    client.use_app(app_id)
    info(f"Removing all WAF whitelists in app, app id: {app_id}")
    whitelists = client.get_all_waf_whitelists()
    for whitelist in whitelists:
        if 'id' in whitelist:
            try:
                info(f"Removing WAF whitelist from app, app id: {app_id}, whitelist id: {whitelist['id']}")
                client.del_waf_whitelist(whitelist['id'])
            except Exception as e:
                error(f"Failed to add WAF whitelist to app, app id: {app_id}, file: {filename}, line: {line(whitelist)}", e)

def check_waf_whitelist(filename, whitelists, ctx):
    client = ctx['client']

    global_waf_rule_sets_map = build_names_map(ctx, "global_waf_rule_sets_map", client.get_all_global_waf_rules)

    for whitelist in whitelists:
        if 'conditions' not in whitelist or not isinstance(whitelist['conditions'], list):
            error(f"Missing or invalid 'conditions' in WAF whitelist, filename: {filename}, line: {line(whitelist)}")

        if 'rules' not in whitelist or not isinstance(whitelist['rules'], list):
            error(f"Missing or invalid 'rules' in WAF whitelist, filename: {filename}, line: {line(whitelist)}")

        for condition in whitelist['conditions']:
            if 'var' not in condition:
                error(f"Missing 'var' in condition, filename: {filename}, line: {line(condition)}")
            if 'op' not in condition:
                error(f"Missing 'op' in condition, filename: {filename}, line: {line(condition)}")
            if 'vals' not in condition and 'val' not in condition:
                error(f"Missing 'vals' or 'val' in condition, filename: {filename}, line: {line(condition)}")

        for rule in whitelist['rules']:
            if 'rule_set_id' not in rule and 'rule_set_name' not in rule:
                error(f"Missing 'rule_set_id' or 'rule_set_name' in rule, filename: {filename}, line: {line(rule)}")

            if 'rule_set_id' in rule and not isinstance(rule['rule_set_id'], int):
                error(f"'rule_set_id' should be an integer, filename: {filename}, line: {line(rule)}")

            if 'rule_set_name' in rule and rule['rule_set_name'] not in global_waf_rule_sets_map:
                error(f"Invalid 'rule_set_name' in rule, filename: {filename}, line: {line(rule)}")

            if 'rule_names' in rule and not isinstance(rule['rule_names'], list):
                error(f"Invalid 'rule_names' in rule, filename: {filename}, line: {line(rule)}")

            if 'rule_names' in rule:
                for name in rule['rule_names']:
                    if not isinstance(name, str):
                        error(f"'rule_name' should be a string, filename: {filename}, line: {line(rule)}")

    return True

def process_waf_whitelist(ctx):
    client = ctx['client']
    app_id = ctx['app_id']
    configs_path = ctx['configs_path']
    location = ctx.get('location', None)

    if not location:
        location = "waf_whitelist"

    configs = read_yaml_config(configs_path, location)
    if not configs:
        return

    info("Checking if WAF whitelists are valid")

    client.use_app(app_id)

    sorted_filenames = sorted(configs.keys())

    for filename in sorted_filenames:
        config = configs[filename]
        check_waf_whitelist(filename, config, ctx)

    # Delete all existing whitelists
    delete_waf_whitelist(client, app_id)

    global_waf_rule_sets_map = ctx.get("global_waf_rule_sets_map", {})

    # Add new whitelists
    for filename in sorted_filenames:
        whitelists = configs[filename]
        # reverse whitelists so that the most recent one is processed first
        whitelists.reverse()
        for whitelist in whitelists:
            conditions = whitelist['conditions']
            rules = whitelist['rules']

            for rule in rules:
                # convert rule_set_name to rule_set_id
                if 'rule_set_name' in rule and rule['rule_set_name'] in global_waf_rule_sets_map:
                    rule['rule_set_id'] = global_waf_rule_sets_map[rule['rule_set_name']]

            try:
                info(f"Adding WAF whitelist to app, app id: {app_id}, file: {filename}, line: {line(whitelist)}")
                client.new_waf_whitelist(condition=conditions, rules=rules)
            except Exception as e:
                error(f"Failed to add WAF whitelist to app, app id: {app_id}, file: {filename}, line: {line(whitelist)}", e)

def cleanup_waf_whitelist(ctx):
    client = ctx['client']
    partition_id = ctx['partition_id']
    domain = ctx['domain']

    app_id = get_app_id_by_domain(client, partition_id, domain)
    if not app_id:
        error(f"App not found, app id: {app_id}, domain: {domain}")

    delete_waf_whitelist(client, app_id)

def export_waf_whitelist(ctx):
    client = ctx['client']
    partition_id = ctx['partition_id']
    domain = ctx['domain']
    configs_path = ctx['export_to_path']

    if ctx.get('app_not_found', False) is True:
        return

    app_id = ctx.get('app_id', None)
    if app_id is None:
        app_id = get_app_id_by_domain(client, partition_id, domain)
        if not app_id:
            warn(f"App not found, partition_id: {partition_id}, domain: {domain}")
            return

        ctx['app_id'] = app_id

    client.use_app(app_id)
    whitelists = client.get_all_waf_whitelists()
    waf_rule_sets = client.get_all_global_waf_rules()

    waf_rule_sets_map = {}
    for waf_rule_set in waf_rule_sets:
        waf_rule_sets_map[waf_rule_set['id']] = waf_rule_set['name']

    if not whitelists:
        info(f"No WAF whitelists found for app_id: {app_id}")
        return

    formatted_whitelists = []
    for whitelist in whitelists:
        formatted_whitelist = {
            'rules': []
        }

        if 'conditions' in whitelist:
            formatted_whitelist['conditions'] = export_conditions(ctx, whitelist['conditions'])

        if 'rules' in whitelist:
            for rule in whitelist['rules']:
                formatted_rule = {
                    'rule_set_name': waf_rule_sets_map[rule['rule_set_id']]
                }

                if 'rule_names' in rule:
                    formatted_rule['rule_names'] = rule['rule_names']
                formatted_whitelist['rules'].append(formatted_rule)

        if 'rule_sets' in whitelist:
            for rule_set_id in whitelist['rule_sets']:
                formatted_rule = {
                    'rule_set_name': waf_rule_sets_map[rule_set_id]
                }
                formatted_whitelist['rules'].append(formatted_rule)

        formatted_whitelists.append(formatted_whitelist)

    export_path = os.path.join(configs_path, 'waf_whitelist')

    try:
        write_yaml_config(export_path, "waf_whitelist.yaml", formatted_whitelists)
        info(f"WAF whitelists exported successfully to waf_whitelist/rules.yaml")
    except Exception as e:
        error(f"Failed to export WAF whitelists to waf_whitelist/rules.yaml", e)



import os
import sys
import re

if sys.argv[1]:
    apps_dir = sys.argv[1]

p = re.compile(r'.*\{\{.+?\}\}.*')

if os.path.isdir(apps_dir):
    for root, dirs, files in os.walk(apps_dir):
        for name in files:
            if name.endswith('.com.conf'):
                app_file = os.path.join(root, name)
                print('include vhost/' + name + ';')
                content = None
                with open(app_file) as f:
                    content = f.read()
                    content = p.sub('', content)
                    content = content.replace('location =/', 'location = /')
                with open(app_file, 'w') as f:
                    f.write(content)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import with_statement

from setuptools import setup, find_packages

import os
import sys

if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    sys.exit()

version = {}
with open(os.path.join(os.path.dirname(__file__),
          "edge2client/version.py")) as fp:
    exec(fp.read(), version)

long_description=""
with open('README.md', 'r') as fp:
    long_description = fp.read()

setup(
    name='openresty-edge-sdk',
    version=version['__version__'],
    description='OpenResty Edge Python SDK',
    platforms='Platform Independent',
    author='OpenResty Inc.',
    author_email='support@openresty.com',
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords=['openresty-edge-sdk', 'edge2client', 'python', 'sdk'],
    url="https://www.openresty.com",
    install_requires=['requests','urllib3','pyOpenSSL','ruamel.yaml','jinja2'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    packages=["edge2client", "edge2client.utils", "edge2yaml"],
    scripts=['bin/edge-config', 'bin/edge2yaml']
)



# -*- coding: utf-8 -*-
# import io
import sdk_test


class TestGlobalK8S(sdk_test.TestSdk):
    def test_global_k8s(self):
        client = self.client

        k8s_id = client.new_global_k8s(name = "k8s_sdk_test_domain", domain = "test.edge-k8s.com",
            port = 9443, token = "eyJhbGciOiJSUzI1NiIsImtpZCI6IiJ9.eyJpc3MiOiJrdWJlcm5ldGVzL3NlcnZpY2VhY2NvdW50Iiwia3ViZXJuZXRlcy5pby9zZXJ2aWNlYWNjb3VudC9uYW1lc3BhY2UiOiJkZWZhdWx0Iiwia3ViZXJuZXRlcy5pby9zZXJ2aWNlYWNjb3VudC9zZWNyZXQubmFtZSI6Im9wZW5yZXN0eS1lZGdlLXd5LXRva2VuLTg1YjVkIiwia3ViZXJuZXRlcy5pby9zZXJ2aWNlYWNjb3VudC9zZXJ2aWNlLWFjY291bnQubmFtZSI6Im9wZW5yZXN0eS1lZGdlLXd5Iiwia3ViZXJuZXRlcy5pby9zZXJ2aWNlYWNjb3VudC9zZXJ2aWNlLWFjY291bnQudWlkIjoiZGRkZjRmODItODcyZi00ODRiLWI1ZTgtYjEwMWZjYzI5ODcxIiwic3ViIjoic3lzdGVtOnNlcnZpY2VhY2NvdW50OmRlZmF1bHQ6b3BlbnJlc3R5LWVkZ2Utd3kifQ.WldqXAWALlSJlo9lY6wsES4iE-BJcEnDSK_YD4M2xMx8qYvrzCdzjcRX6Dl56tR1Yv98QutEcyG_hjfjdYt_pDI_pgni3f4-jEjXrSDOiPw7kXschhaqVaN1mDEOGuOi2mtGtfEz6tPg-j4dbUuvmR3S5niBAh9tbe9_IacRiEt9HEYw0W-knoHyrOofHUY99JCnVjtXADhfLMQQyo0Ok8FNEjb06qMWVF_vwy6CJKWZy6dmVX1SXT494rToy97dEvfHtMqSxQOnCEKZ-8XwxgUTBPExpbX52tyJ8D2xmIorQVnJovCL2anwLMieYkD5F11KtFDozI510AyXR2i9FA", ssl_verify = False)
 
        self.assertIs(type(k8s_id), int)
        self.assertGreater(k8s_id, 0)
        self.global_k8s_id = k8s_id

        data = client.get_global_k8s(k8s_id)

        self.assertEqual(data["name"], "k8s_sdk_test_domain")
        self.assertEqual(data["domain"], "test.edge-k8s.com")
        self.assertEqual(data["port"], 9443)
        self.assertEqual(data["id"], k8s_id)
        self.assertEqual(data["ssl_verify"], False)

        all_k8s = client.get_all_global_k8s()

        self.assertEqual(len(all_k8s), 1)

        self.assertEqual(all_k8s[0]["id"], k8s_id)

        client.del_global_k8s(k8s_id)




import hashlib
import sys
import traceback
import ipaddress
import json
import re
import os
from OpenSSL import crypto

def error(msg, e=None):
    if isinstance(msg, str):
        msg = msg[0].upper() + msg[1:]

    if e:
        stack_trace = traceback.format_exc()
        sys.exit(f'[!] Error: {msg}, detail: {stack_trace}.')
    else:
        sys.stderr.write(f'[!] Error: {msg}, detail:\n')
        stack_trace = traceback.extract_stack()
        for frame in stack_trace:
            sys.stderr.write(f"  File \"{frame.filename}\", line {frame.lineno}, in {frame.name}\n")
        sys.exit(1)

def warn(msg):
    if isinstance(msg, str):
        msg = msg[0].upper() + msg[1:]

    print(f'[?] Warn: {msg}.')

def info(msg):
    if isinstance(msg, str):
        msg = msg[0].upper() + msg[1:]

    print(f'[+] Info: {msg}.')

def md5sum(string_to_hash):
    hash_object = hashlib.md5()
    hash_object.update(string_to_hash.encode('utf-8'))
    md5_hash = hash_object.hexdigest()
    return md5_hash

def cal_config_md5(config):
    json_str = json.dumps(config, ensure_ascii=False, sort_keys=True)
    return md5sum(json_str)

def to_fake_name(name, partition_id, sync_to_all):
    if sync_to_all:
        return name

    return f"{name}_for_partition_{partition_id}"

def to_real_name(name, partition_id, sync_to_all):
    if sync_to_all:
        return name

    suffix = f"_for_partition_{partition_id}"
    if suffix in name:
        return name.split(suffix)[0]

    return None

def extract_name(name):
    pattern = r'^(.+)_for_partition_\d+$'

    match = re.match(pattern, name)
    if match:
        return match.group(1)
    else:
        return name

def is_partition_action(name, partition_id):
    suffix = f"_for_partition_{partition_id}"
    return name.endswith(suffix)

def generate_certificates(domains, domain):
    # 1. Generate the CA Root certificate and key
    ca_key = crypto.PKey()
    ca_key.generate_key(crypto.TYPE_RSA, 4096)

    ca_cert = crypto.X509()
    ca_cert.set_version(2)
    ca_cert.set_serial_number(1)
    ca_subject = ca_cert.get_subject()
    ca_subject.CN = f"{domain}.ca"
    ca_subject.C = "US"
    ca_subject.L = "San Fransisco"
    ca_cert.set_issuer(ca_subject)
    ca_cert.set_pubkey(ca_key)
    ca_cert.gmtime_adj_notBefore(0)
    ca_cert.gmtime_adj_notAfter(3650 * 24 * 60 * 60)
    ca_cert.sign(ca_key, "sha256")

    rootCA_key = crypto.dump_privatekey(crypto.FILETYPE_PEM, ca_key)
    rootCA_crt = crypto.dump_certificate(crypto.FILETYPE_PEM, ca_cert)

    # 2. Generate the server private key
    server_key = crypto.PKey()
    server_key.generate_key(crypto.TYPE_RSA, 2048)
    edge_node_key = crypto.dump_privatekey(crypto.FILETYPE_PEM, server_key)

    # 3. Create the certificate signing request (csr)
    server_csr = crypto.X509Req()
    server_subject = server_csr.get_subject()
    server_subject.C = "US"
    server_subject.ST = "CA"
    server_subject.O = "Organization, Inc."
    server_subject.CN = f"{domain}"

    excluded_values = [ domain ]
    # Add Subject Alternative Names (SAN) for additional domains
    san_list = [f"DNS:{dn}" for dn in domains if dn not in excluded_values]
    if san_list:
        san_ext = crypto.X509Extension(b"subjectAltName", False, ", ".join(san_list).encode())
        server_csr.add_extensions([san_ext])

    server_csr.set_pubkey(server_key)
    server_csr.sign(server_key, "sha256")
    edge_node_csr = crypto.dump_certificate_request(crypto.FILETYPE_PEM, server_csr)

    # 4. Generate the certificate using the csr and key along with the CA Root key
    server_cert = crypto.X509()
    server_cert.set_serial_number(1)
    server_cert.gmtime_adj_notBefore(0)
    server_cert.gmtime_adj_notAfter(3650 * 24 * 60 * 60)
    server_cert.set_issuer(ca_cert.get_subject())
    server_cert.set_subject(server_csr.get_subject())
    server_cert.set_pubkey(server_csr.get_pubkey())
    if san_list:
        server_cert.add_extensions(server_csr.get_extensions())
    server_cert.sign(ca_key, "sha256")
    edge_node_crt = crypto.dump_certificate(crypto.FILETYPE_PEM, server_cert)

    rootCA_key = rootCA_key.decode('utf-8')
    rootCA_crt = rootCA_crt.decode('utf-8')
    edge_node_key = edge_node_key.decode('utf-8')
    edge_node_csr = edge_node_csr.decode('utf-8')
    edge_node_crt = edge_node_crt.decode('utf-8')

    return rootCA_key, rootCA_crt, edge_node_key, edge_node_csr, edge_node_crt

def release_changes(client, app_id, staging_release=None):
    try:
        if client.pending_changes(app_id=app_id) > 0:
            if staging_release:
                info(f"staging release pending changes of app id: {app_id}")
            else:
                info(f"releasing pending changes of app id: {app_id}")
            client.new_release(app_id=app_id, gray=staging_release)
    except Exception as e:
        error(f"releasing pending changes of app failed, app id: {app_id}", e)

    if not client.request_ok():
        error(f"clear pending changes of app failed, app id: {app_id}")

def clear_changes(client, app_id):
    try:
        if client.pending_changes(app_id=app_id) > 0:
            info(f"clearing pending changes of app id: {app_id}")
            client.clear_pending_changes(app_id=app_id)
    except Exception as e:
        error(f"clear pending changes of app failed, app id: {app_id}", e)

    if not client.request_ok():
        error(f"clear pending changes of app failed, app id: {app_id}")

def release_partition_changes(client, partition_id):
    try:
        count = client.pending_changes(app_type="partition", app_id=partition_id)
        if count > 0:
            info(f"releasing pending changes of partition id: {partition_id}")
            client.new_release(app_type="partition", app_id=partition_id)
    except Exception as e:
        warn(f"releasing pending changes of partition failed, partition id: {partition_id}, error: {e}")
        return

def clear_partition_changes(client, partition_id):
    try:
        count = client.pending_changes(app_type="partition", app_id=partition_id)
        if count > 0:
            info(f"clearing pending changes of partition id: {partition_id}")
            client.clear_pending_changes(app_type="partition", app_id=partition_id)
    except Exception as e:
        warn(f"clearing pending changes of partition failed, partition id: {partition_id}, error: {e}")
        return

def fix_line(l):
    if isinstance(l, int):
        return l + 1
    return None

def is_yaml_class(obj):
    if str(type(obj)).startswith("<class 'ruamel.yaml."):
        return True

    return False

def line(var):
    if is_yaml_class(var):
        return var.lc.line + 1

    if isinstance(var, dict):
        for k, v in var.items():
            if is_yaml_class(v):
                return v.lc.line + 1

        return None
    elif isinstance(var, list):
        for v in var:
            if is_yaml_class(v):
                return v.lc.line + 1

        return None

    return None

def check_type(value, expected_type):
    if isinstance(expected_type, type):
        expected_type = [expected_type]

    for t in expected_type:
        if isinstance(value, t):
            return True

    return False

def get_real_comment(comment):
    if not comment:
        return ''

    pattern = r"md5: [a-f0-9A-F]+, please do not modify\."
    comment = re.sub(pattern, "", comment)
    return comment.strip()

def get_md5_from_comment(comment):
    if not comment:
        return None

    match = re.search(r'md5: ([a-f0-9A-F]+), please do not modify\.', comment)
    if match:
        return match.group(1)

    return None

def get_updated_list(new_dict, old_list, major_key, compare_keys, changed_only=False):
    # check if config changed
    need_update = False
    updated_list = list()
    for old_config in old_list:
        if major_key in old_config and old_config[major_key] in new_dict:
            major = old_config[major_key]
            new_config = new_dict[major]
            for compare_key in compare_keys:
                old_val = old_config.get(compare_key, None)
                new_val = new_config.get(compare_key, None)
                if old_val != new_val:
                    old_config[compare_key] = new_val
                    need_update = True
                    if changed_only:
                        updated_list.append(old_config)

            new_dict[major] = None

        if not changed_only:
            updated_list.append(old_config)

    # new format
    for key, new_config in new_dict.items():
        if new_config is not None:
            updated_list.append(new_config)
            need_update = True

    if need_update:
        return updated_list

    return None

def is_valid_ipv4_address(address):
    try:
        ipaddress.ip_address(address)
        return True
    except ValueError:
        return False

def is_valid_ipv6_address(address):
    try:
        ipaddress.IPv6Address(address)
        return True
    except ValueError:
        return False

def is_valid_hostname(hostname):
    if len(hostname) > 255:
        return False
    if hostname[-1] == ".":
        hostname = hostname[:-1]  # strip exactly one dot from the right, if present
    allowed = re.compile("(?!-)[-A-Z0-9]{1,63}(?<!-)$", re.IGNORECASE)
    return all(allowed.match(x) for x in hostname.split("."))

def is_valid_host(host):
    return is_valid_ipv4_address(host) or is_valid_ipv6_address(host) or is_valid_hostname(host)

def get_app_id_by_domain(client, partition_id, domain):
    uri = f"applications?detail=1&type=http&partition_id={partition_id}"
    apps = client.get_all(uri, True)
    for app in apps:
        for dm in app['domains']:
            if dm == domain:
                return app['id']

    return None

def get_app_config_by_id(client, app_id):
    return
    uri = f"applications?detail=1&partition_id={partition_id}"
    apps = client.get_all(uri, True)
    for app in apps:
        for dm in app['domains']:
            if dm['domain'] == domain:
                return app['id']

    return None

if __name__ == "__main__":
    rootCA_key, rootCA_crt, edge_node_key, edge_node_csr, edge_node_crt = generate_certificates(["test.com", "www.test.com"], "test.com")
    print(f"rootCA_key: {rootCA_key}\nrootCA_crt: {rootCA_crt}\nedge_node_key: {edge_node_key}\nedge_node_csr: {edge_node_csr}\n, edge_node_crt: {edge_node_crt}\n")



# -*- coding: utf-8 -*-
# import io
import sdk_test
from datetime import datetime

class TestSearch(sdk_test.TestSdk):
    def test_search_app_by_upstream_ip(self):
        app_id = self.app_id

        up_id = self.client.new_upstream(
            name='origin-upstream',
            servers=[
                {'ip': '172.22.31.1', 'port': 80},
                {'ip': '172.22.31.2', 'port': 80, 'weight': 2}
            ])
        self.assertIs(type(up_id), int)
        self.assertGreater(up_id, 0)

        changes = self.client.pending_changes()
        self.assertEqual(changes, 1)

        new_app_id = self.client.new_app(domains=['prefix' + self.apex],
                                         label=self.apex)
        self.assertIs(type(app_id), int)
        self.assertGreater(app_id, 0)

        self.client.use_app(new_app_id)

        up_id = self.client.new_upstream(
            name='origin-upstream',
            servers=[
                {'ip': '172.22.31.1', 'port': 80},
                {'ip': '172.22.31.3', 'port': 80, 'weight': 2}
            ])
        self.assertIs(type(up_id), int)
        self.assertGreater(up_id, 0)

        data = self.client.search_app(upstream_ip='172.22.31.1')
        ok = self.client.del_app(new_app_id)
        self.assertTrue(ok)
        self.client.use_app(app_id)

        self.assertEqual(len(data), 2)
        self.assertEqual(data[1]['app']['domains'][0]['domain'], self.apex)
        self.assertEqual(data[1]['app']['id'], app_id)
        self.assertEqual(data[0]['app']['domains'][0]['domain'], 'prefix' + self.apex)
        self.assertEqual(data[0]['app']['id'], new_app_id)

        ok = self.client.del_upstream(up_id)
        self.assertTrue(ok)

    def test_req_rewrite(self):
        condition = [
            {'var': 'host', 'val': 'con.' + self.apex},
            {'var': ['req-header', 'Referer'],
             'vals': [[r'foo\d+', 'rx'], 'foo.com']}
        ]
        conseq = {
            'enable-websocket': {},
            'redirect': {'url': '/cn/2017/', 'code': 302}
        }
        rule_id = self.client.new_rule(condition=condition, conseq=conseq)
        self.assertIs(type(rule_id), int)
        self.assertGreater(rule_id, 0)

        data = self.client.get_rule(rule_id)
        if data['actions'][0]['type'] == 'redirect':
            self.assertEqual(data['actions'][0]['redirect']['url'],
                             '/cn/2017/')
            self.assertEqual(data['actions'][1]['type'], 'enable-websocket')
        else:
            self.assertEqual(data['actions'][1]['redirect']['url'],
                             '/cn/2017/')
            self.assertEqual(data['actions'][0]['type'], 'enable-websocket')

        conseq['redirect']['url'] = '/cn/2018/'
        ok = self.client.put_rule(
            rule_id=rule_id, condition=condition, conseq=conseq)
        self.assertTrue(ok)
        data = self.client.get_rule(rule_id)
        if data['actions'][0]['type'] == 'redirect':
            self.assertEqual(data['actions'][0]['redirect']['url'],
                             '/cn/2018/')
        else:
            self.assertEqual(data['actions'][1]['redirect']['url'],
                             '/cn/2018/')

        ok = self.client.del_rule(rule_id)
        self.assertTrue(ok)
        data = self.client.get_rule(rule_id)
        self.assertEqual(data, {})


    def test_search_upstream(self):
        app_id = self.app_id

        up_id = self.client.new_upstream(
            name='app-upstream-' + datetime.now().strftime("%s"),
            servers=[
                {'ip': '172.22.31.1', 'port': 80},
                {'ip': '172.22.31.2', 'port': 80, 'weight': 2}
            ])
        self.assertIs(type(up_id), int)
        self.assertGreater(up_id, 0)

        changes = self.client.pending_changes()
        self.assertEqual(changes, 1)

        new_app_id = self.client.new_app(domains=['prefix' + self.apex],
                                         label=self.apex)
        self.assertIs(type(app_id), int)
        self.assertGreater(app_id, 0)

        global_up_id = self.client.new_global_upstream(
            name='global-upstream-' + datetime.now().strftime("%s"),
            servers=[
                {'ip': '172.22.31.1', 'port': 80},
                {'ip': '172.22.31.3', 'port': 80, 'weight': 2}
            ])

        self.assertIs(type(global_up_id), int)

        data = self.client.search_upstream_by_ip("172.22.31.1", 1, 1)
        self.assertEqual(len(data), 1)

        data = self.client.search_upstream_by_ip("172.22.31.1")
        self.assertEqual(len(data), 2)

        data = self.client.search_upstream_by_ip("172.22.31.2")
        self.assertEqual(len(data), 1)
        self.assertIn("app-upstream", data[0]['name'])

        data = self.client.search_upstream_by_ip("172.22.31.3")
        self.assertEqual(len(data), 1)
        self.assertIn("global-upstream", data[0]['name'])

        data = self.client.search_upstream_by_name("app-upstream")
        self.assertEqual(len(data), 1)
        self.assertIn("app-upstream", data[0]['name'])

        data = self.client.search_upstream_by_name("global-upstream")
        self.assertEqual(len(data), 1)
        self.assertIn("global-upstream", data[0]['name'])

        ok = self.client.del_global_upstream(global_up_id)
        self.assertTrue(ok)

        ok = self.client.del_upstream(up_id)
        self.assertTrue(ok)


    def test_search_http_app(self):

        new_app_id1 = self.client.new_app(domains=[datetime.now().strftime("%s") + ".domain.com"],
                                         label="label-name-" + datetime.now().strftime("%s"))
        self.assertIs(type(new_app_id1), int)
        self.assertGreater(new_app_id1, 0)

        new_app_id2 = self.client.new_app(domains=[datetime.now().strftime("%s") + ".domain2.com"],
                                         label="label2-name-" + datetime.now().strftime("%s"))
        self.assertIs(type(new_app_id2), int)
        self.assertGreater(new_app_id2, 0)

        data = self.client.search_http_app_by_keyword("domain", 1, 1)
        self.assertEqual(len(data), 1)

        data = self.client.search_http_app_by_keyword("domain")
        self.assertEqual(len(data), 2)

        data = self.client.search_http_app_by_keyword("domain2")
        self.assertEqual(len(data), 1)
        self.assertIn("label2-name", data[0]['name'])

        data = self.client.search_http_app_by_keyword("label-name")
        self.assertEqual(len(data), 1)
        self.assertIn("label-name", data[0]['name'])

        ok = self.client.del_app(new_app_id1)
        self.assertTrue(ok)

        ok = self.client.del_app(new_app_id2)
        self.assertTrue(ok)



#!/usr/bin/python
# -*- coding: UTF-8 -*-
"""
This file is used to release app
"""

import time

try:
    import py_client
except ImportError:
    print("Error: cannot import py_client.py")
    exit(1)


def main():
    client = py_client.get_client()
    app_ids = client.get_all_apps(detail=False)

    for app_id in app_ids:
        client.use_app(app_id)
        ret = client.new_release()
        if ret:
            print('release ok app_id: ' + str(app_id))
        else:
            print('release failed: ' + str(app_id))
        time.sleep(0.1)



if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# import io
import sdk_test


class TestVersion(sdk_test.TestSdk):
    def test_get_version(self):
        data=self.client.get_version()
        self.assertGreater(data['admin_db_version'], 0)
        self.assertGreater(data['nginx_version'], 0)
        self.assertGreater(data['ngx_lua_version'], 0)
        self.assertGreater(data['required'], 0)
        self.assertGreater(data['version'], 0)



import os

from jinja2 import Template

from .utils import error, warn, info, line
from .read_config import read_yaml_config, write_yaml_config
from .edge_email import send_email

login_types = {
    'Normal': 'normal',
    'LDAP': 'ldap',
    'OpenIDC': 'oidc',
    'none': 'none',
}

login_type_names = {
    'normal': 'Normal',
    'ldap': 'LDAP',
    'oidc': 'OpenIDC',
    'none': 'none',
}

def render_email(template, username, password, user_id, user_group, user_group_id):
    tm = Template(template)
    return tm.render(username=username, password=password,
                     user_id=user_id, user_group=user_group,
                     user_group_id=user_group_id)

def send_email_to_user(email_configs, user, user_id, group_names, group_ids):
    subject = email_configs['new_user_email_template']['subject']
    content = email_configs['new_user_email_template']['content']
    content = render_email(content, user['username'], user['password'],
                            user_id, ",".join(group_names), ",".join(map(str, group_ids)))
    info("sending email to user: " + user['username'] + ", email: " + user['email'])
    ok = send_email(email_configs, user['email'], subject, content)
    if not ok:
        error("send email to " + user['email'] + " failed")

def process_users(ctx):
    client = ctx['client']
    configs_path = ctx['users_config_path']
    email_configs = ctx['email_configs']

    if configs_path is None:
        return True

    configs = read_yaml_config(configs_path)
    if configs is None:
        return True

    users = list()
    user_groups = None
    sorted_filenames = sorted(configs.keys())
    # check users and format users
    for filename in sorted_filenames:
        config = configs[filename]
        if not isinstance(config, dict):
            error(f"unsupported file format for users, file: {filename}")

        if 'users' not in config:
            error(f"users not found in file: {filename}")

        def_change_pwd_on_login = config.get('change_pwd_on_login', False)
        def_login_type = config.get('login_type', 'Normal')
        def_password = config.get('password', None)
        def_allow_login = config.get('allow_login', True)
        def_group_name = config.get('group', None)
        def_send_email_to_existing_user = config.get('send_email_to_existing_user', False)
        def_send_email_to_new_user = config.get('send_email_to_new_user', False)

        for user in config['users']:
            if not isinstance(user, dict):
                error(f"unsupported file format for users, file: {filename}, line: {line(user)}")

            if 'username' not in user:
                error(f"username not found in user, file: {filename}, line: {line(user)}")

            login_type = user.get('login_type', None)
            if login_type is None:
                warn(f"Login type not found in user, file: {filename}, line: {line(user)}, using default login type: {def_login_type}")
                login_type = [ def_login_type ]

            if not isinstance(login_type, list):
                login_type = [ login_type ]

            for typ in login_type:
                if typ not in login_types:
                    error(f"unsupported login type: {typ}, file: {filename}, line: {line(user)}")

            password = user.get('password', def_password)
            if not isinstance(password, str):
                error(f"password must be a string, file: {filename}, line: {line(user)}")

            allow_login = user.get('allow_login', def_allow_login)
            if not isinstance(allow_login, bool):
                error(f"allow_login must be a boolean, file: {filename}, line: {line(user)}")

            if allow_login is False:
                login_type = [ 'none' ]

            change_pwd_on_login = user.get('change_pwd_on_login', def_change_pwd_on_login)
            if not isinstance(change_pwd_on_login, bool):
                error(f"change_pwd_on_login must be a boolean, file: {filename}, line: {line(user)}")

            group_name = user.get('group', def_group_name)
            if not isinstance(group_name, str) and not isinstance(group_name, list):
                error(f"group must be a string, file: {filename}, line: {line(user)}")

            if isinstance(group_name, str):
                group_names = [ group_name ]

            # check if user group exists
            if user_groups is None:
                groups = client.get_all_user_groups()
                if groups is None:
                    error(f"failed to get user groups")

                user_groups = dict()
                for g in groups:
                    # g['permission'] = None
                    name = g.get('group_name')
                    user_groups[name] = g['id']

            for group in group_names:
                if group not in user_groups:
                    error(f"user group not found: {group}, file: {filename}, line: {line(user)}")

            if 'email' not in user and user.get('send_email_to_new_user', False) == True:
                error(f"email not found in user, file: {filename}, line: {line(user)}")

            real_login_types = [ login_types[typ] for typ in login_type ]
            real_gids = [ user_groups[group] for group in group_names ]

            users.append({
                'username': user['username'],
                'password': password,
                'requires_password_change': change_pwd_on_login,
                'login_type': real_login_types,
                'gid': real_gids,
                'group_names': group_names,
                'email': user.get('email', None),
                'send_email_to_existing_user': user.get('send_email_to_existing_user', def_send_email_to_existing_user),
                'send_email_to_new_user': user.get('send_email_to_new_user', def_send_email_to_new_user),
            })

    for user in users:
        # check if user exists
        data = client.search_global_user(user['username'])
        user_id = data.get('id', None)

        if isinstance(user_id, int):
            # check next users
            warn(f"User {user['username']} already exists, id: {user_id}")

            if 'email' in user and user.get('send_email_to_existing_user', None) == True:
                if email_configs is None:
                    warn(f"unable to send email to {user['email']}, because the sender's email config was not found")
                else:
                    send_email_to_user(email_configs, user, user_id, user['group_names'], user['gid'])

            continue

        # new user
        data = client.add_global_user(user['username'], user['password'],
                               gid=user['gid'],
                               login_type=user['login_type'],
                               requires_password_change=user['requires_password_change'])

        user_id = data.get('id', None)
        if not isinstance(user_id, int):
            error(f"failed to add user: {user['username']}")

        info("added user: " + user['username'] + ", id: " + str(user_id) + ", user groups: " + ",".join(user['group_names']))

        # send email
        if 'email' in user and user.get('send_email_to_new_user', False) == True:
            if email_configs is None:
                warn(f"unable to send email to {user['email']}, because the sender's email config was not found")
            else:
                send_email_to_user(email_configs, user, user_id, user['group_names'], user_groups)

def export_users_and_groups(ctx):
    client = ctx['client']
    configs_path = ctx['export_to_path']
    export_users = ctx.get('export_users', False)

    # Export user groups and their users
    info("Exporting user groups and users...")

    if export_users is False:
        warn("User passwords will not be exported")

    export_path = os.path.join(configs_path, "users")

    user_groups = client.get_all_user_groups()
    groups_map = dict()
    for group in user_groups:
        group_name = group['group_name']
        group_id = group['id']
        groups_map[group_id] = group_name

    users = client.get_all_global_users(detail=True)
    group_users = dict()
    for user in users:
        # for gid in user['gid']:
        # just add to the first group
        gid = user['gid'][0]
        group_name = groups_map.get(gid)
        if group_name not in group_users:
            group_users[group_name] = list()

        group_users[group_name].append(user)

    for group in user_groups:
        group_name = group['group_name']
        group_id = group['id']

        export_data = {
            'group': group_name,
            'users': []
        }

        if export_users is True:
            users = group_users.get(group_name, list())
            for user in users:
                # user_details = client.get_global_user(user['id'])
                formatted_user = {
                    'username': user['username'],
                    'email': user.get('email', None),
                    'group': [ groups_map.get(gid) for gid in user['gid'] ],
                    'login_type': [ login_type_names[name] for name in user.get('login_type', ['normal']) ],
                    'allow_login': user.get('login_type', ['normal'])[0] != 'none',
                    'change_pwd_on_login': user.get('requires_password_change', False),
                    'send_email_to_existing_user': False,  # Default value, adjust as needed
                    'send_email_to_new_user': False,  # Default value, adjust as needed
                }
                export_data['users'].append(formatted_user)

        # Write to YAML file
        try:
            write_yaml_config(export_path, f"{group_name}.yaml", export_data)
            if export_users is True:
                info(f"User group '{group_name}' and its users exported successfully to users/{group_name}.yaml")
            else:
                info(f"User group '{group_name}' exported successfully to users/{group_name}.yaml")
        except Exception as e:
            error(f"Failed to export user group '{group_name}' and its users to users/{group_name}.yaml", e)

    info(f"All user groups and users exported successfully to users/")
